"""Module for defining various general equations."""

from typing import Callable
import torch
from .equation import Equation
from ..operator import grad, div, laplacian
from ..utils import check_consistency

# Pylint warning disabled because the classes defined in this module
# inherit from Equation and are meant to be simple containers for equations.


class FixedValue(Equation):  # pylint: disable=R0903
    """
    Equation to enforce a fixed value. Can be used to enforce Dirichlet Boundary
    conditions.
    """

    def __init__(self, value, components=None):
        """
        Initialization of the :class:`FixedValue` class.

        :param float value: The fixed value to be enforced.
        :param list[str] components: The name of the output variables for which
            the fixed value condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        """

        def equation(_, output_):
            """
            Definition of the equation to enforce a fixed value.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            if components is None:
                return output_ - value
            return output_.extract(components) - value

        super().__init__(equation)


class FixedGradient(Equation):  # pylint: disable=R0903
    """
    Equation to enforce a fixed gradient for a specific condition.
    """

    def __init__(self, value, components=None, d=None):
        """
        Initialization of the :class:`FixedGradient` class.

        :param float value: The fixed value to be enforced to the gradient.
        :param list[str] components: The name of the output variables for which
            the fixed gradient condition is applied. It should be a subset of
            the output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the
            gradient is computed. It should be a subset of the input labels.
            If ``None``, all the input variables are considered.
            Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a fixed gradient.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            return grad(output_, input_, components=components, d=d) - value

        super().__init__(equation)


class FixedFlux(Equation):  # pylint: disable=R0903
    """
    Equation to enforce a fixed flux, or divergence, for a specific condition.
    """

    def __init__(self, value, components=None, d=None):
        """
        Initialization of the :class:`FixedFlux` class.

        :param float value: The fixed value to be enforced to the flux.
        :param list[str] components: The name of the output variables for which
            the fixed flux condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the flux
            is computed. It should be a subset of the input labels. If ``None``,
            all the input variables are considered. Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a fixed flux.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            return div(output_, input_, components=components, d=d) - value

        super().__init__(equation)


class FixedLaplacian(Equation):  # pylint: disable=R0903
    """
    Equation to enforce a fixed laplacian for a specific condition.
    """

    def __init__(self, value, components=None, d=None):
        """
        Initialization of the :class:`FixedLaplacian` class.

        :param float value: The fixed value to be enforced to the laplacian.
        :param list[str] components: The name of the output variables for which
            the fixed laplace condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the
            laplacian is computed. It should be a subset of the input labels.
            If ``None``, all the input variables are considered.
            Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a fixed laplacian.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            return (
                laplacian(output_, input_, components=components, d=d) - value
            )

        super().__init__(equation)


class Laplace(FixedLaplacian):  # pylint: disable=R0903
    r"""
    Equation to enforce a null laplacian for a specific condition.
    The equation is defined as follows:

    .. math::

        \delta u = 0

    """

    def __init__(self, components=None, d=None):
        """
        Initialization of the :class:`Laplace` class.

        :param list[str] components: The name of the output variables for which
            the null laplace condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the
            laplacian is computed. It should be a subset of the input labels.
            If ``None``, all the input variables are considered.
            Default is ``None``.
        """
        super().__init__(0.0, components=components, d=d)


class Advection(Equation):  # pylint: disable=R0903
    r"""
    Implementation of the N-dimensional advection equation with constant
    velocity parameter. The equation is defined as follows:

    .. math::

        \frac{\partial u}{\partial t} + c \cdot \nabla u = 0

    Here, :math:`c` is the advection velocity parameter.
    """

    def __init__(self, c):
        """
        Initialization of the :class:`Advection` class.

        :param c: The advection velocity. If a scalar is provided, the same
            velocity is applied to all spatial dimensions. If a list is
            provided, it must contain one value per spatial dimension.
        :type c: float | int | List[float] | List[int]
        :raises ValueError: If ``c`` is an empty list.
        """
        # Check consistency
        check_consistency(c, (float, int, list))
        if isinstance(c, list):
            all(check_consistency(ci, (float, int)) for ci in c)
            if len(c) < 1:
                raise ValueError("'c' cannot be an empty list.")
        else:
            c = [c]

        # Store advection velocity parameter
        self.c = torch.tensor(c).unsqueeze(0)

        def equation(input_, output_):
            """
            Implementation of the advection equation.

            :param LabelTensor input_: The input data of the problem.
            :param LabelTensor output_: The output data of the problem.
            :return: The residual of the advection equation.
            :rtype: LabelTensor
            :raises ValueError: If the ``input_`` labels do not contain the time
                variable 't'.
            :raises ValueError: If ``c`` is a list and its length is not
                consistent with the number of spatial dimensions.
            """
            # Store labels
            input_lbl = input_.labels
            spatial_d = [di for di in input_lbl if di != "t"]

            # Ensure time is passed as input
            if "t" not in input_lbl:
                raise ValueError(
                    "The ``input_`` labels must contain the time 't' variable."
                )

            # Ensure consistency of c length
            if self.c.shape[-1] != len(input_lbl) - 1 and self.c.shape[-1] > 1:
                raise ValueError(
                    "If 'c' is passed as a list, its length must be equal to "
                    "the number of spatial dimensions."
                )

            # Repeat c to ensure consistent shape for advection
            c = self.c.repeat(output_.shape[0], 1)
            if c.shape[1] != (len(input_lbl) - 1):
                c = c.repeat(1, len(input_lbl) - 1)

            # Add a dimension to c for the following operations
            c = c.unsqueeze(-1)

            # Compute the time derivative and the spatial gradient
            time_der = grad(output_, input_, components=None, d="t")
            grads = grad(output_=output_, input_=input_, d=spatial_d)

            # Reshape and transpose
            tmp = grads.reshape(*output_.shape, len(spatial_d))
            tmp = tmp.transpose(-1, -2)

            # Compute advection term
            adv = (tmp * c).sum(dim=tmp.tensor.ndim - 2)

            return time_der + adv

        super().__init__(equation)


class AllenCahn(Equation):  # pylint: disable=R0903
    r"""
    Implementation of the N-dimensional Allen-Cahn equation, defined as follows:

    .. math::

        \frac{\partial u}{\partial t} - \alpha \Delta u + \beta(u^3 - u) = 0

    Here, :math:`\alpha` and :math:`\beta` are parameters of the equation.
    """

    def __init__(self, alpha, beta):
        """
        Initialization of the :class:`AllenCahn` class.

        :param alpha: The diffusion coefficient.
        :type alpha: float | int
        :param beta: The reaction coefficient.
        :type beta: float | int
        """
        check_consistency(alpha, (float, int))
        check_consistency(beta, (float, int))
        self.alpha = alpha
        self.beta = beta

        def equation(input_, output_):
            """
            Implementation of the Allen-Cahn equation.

            :param LabelTensor input_: The input data of the problem.
            :param LabelTensor output_: The output data of the problem.
            :return: The residual of the Allen-Cahn equation.
            :rtype: LabelTensor
            :raises ValueError: If the ``input_`` labels do not contain the time
                variable 't'.
            """
            # Ensure time is passed as input
            if "t" not in input_.labels:
                raise ValueError(
                    "The ``input_`` labels must contain the time 't' variable."
                )

            # Compute the time derivative and the spatial laplacian
            u_t = grad(output_, input_, d=["t"])
            u_xx = laplacian(
                output_, input_, d=[di for di in input_.labels if di != "t"]
            )

            return u_t - self.alpha * u_xx + self.beta * (output_**3 - output_)

        super().__init__(equation)


class DiffusionReaction(Equation):  # pylint: disable=R0903
    r"""
    Implementation of the N-dimensional Diffusion-Reaction equation,
    defined as follows:

    .. math::

        \frac{\partial u}{\partial t} - \alpha \Delta u - f = 0

    Here, :math:`\alpha` is a parameter of the equation, while :math:`f` is the
    reaction term.
    """

    def __init__(self, alpha, forcing_term):
        """
        Initialization of the :class:`DiffusionReaction` class.

        :param alpha: The diffusion coefficient.
        :type alpha: float | int
        :param Callable forcing_term: The forcing field function, taking as
            input the points on which evaluation is required.
        """
        check_consistency(alpha, (float, int))
        check_consistency(forcing_term, (Callable))
        self.alpha = alpha
        self.forcing_term = forcing_term

        def equation(input_, output_):
            """
            Implementation of the Diffusion-Reaction equation.

            :param LabelTensor input_: The input data of the problem.
            :param LabelTensor output_: The output data of the problem.
            :return: The residual of the Diffusion-Reaction equation.
            :rtype: LabelTensor
            :raises ValueError: If the ``input_`` labels do not contain the time
                variable 't'.
            """
            # Ensure time is passed as input
            if "t" not in input_.labels:
                raise ValueError(
                    "The ``input_`` labels must contain the time 't' variable."
                )

            # Compute the time derivative and the spatial laplacian
            u_t = grad(output_, input_, d=["t"])
            u_xx = laplacian(
                output_, input_, d=[di for di in input_.labels if di != "t"]
            )

            return u_t - self.alpha * u_xx - self.forcing_term(input_)

        super().__init__(equation)


class Helmholtz(Equation):  # pylint: disable=R0903
    r"""
    Implementation of the Helmholtz equation, defined as follows:

    .. math::

            \Delta u + k u - f = 0

    Here, :math:`k` is a parameter of the equation, while :math:`f` is the
    forcing term.
    """

    def __init__(self, k, forcing_term):
        """
        Initialization of the :class:`Helmholtz` class.

        :param k: The parameter of the equation.
        :type k: float | int
        :param Callable forcing_term: The forcing field function, taking as
            input the points on which evaluation is required.
        """
        check_consistency(k, (int, float))
        check_consistency(forcing_term, (Callable))
        self.k = k
        self.forcing_term = forcing_term

        def equation(input_, output_):
            """
            Implementation of the Helmholtz equation.

            :param LabelTensor input_: The input data of the problem.
            :param LabelTensor output_: The output data of the problem.
            :return: The residual of the Helmholtz equation.
            :rtype: LabelTensor
            """
            lap = laplacian(output_, input_)
            return lap + self.k * output_ - self.forcing_term(input_)

        super().__init__(equation)


class Poisson(Equation):  # pylint: disable=R0903
    r"""
    Implementation of the Poisson equation, defined as follows:

    .. math::

            \Delta u - f = 0

    Here, :math:`f` is the forcing term.
    """

    def __init__(self, forcing_term):
        """
        Initialization of the :class:`Poisson` class.

        :param Callable forcing_term: The forcing field function, taking as
            input the points on which evaluation is required.
        """
        check_consistency(forcing_term, (Callable))
        self.forcing_term = forcing_term

        def equation(input_, output_):
            """
            Implementation of the Poisson equation.

            :param LabelTensor input_: The input data of the problem.
            :param LabelTensor output_: The output data of the problem.
            :return: The residual of the Poisson equation.
            :rtype: LabelTensor
            """
            lap = laplacian(output_, input_)
            return lap - self.forcing_term(input_)

        super().__init__(equation)



"""Module for the Self-Adaptive PINN solver."""

from copy import deepcopy
import torch

from ...utils import check_consistency
from ...problem import InverseProblem
from ..solver import MultiSolverInterface
from .pinn_interface import PINNInterface


class Weights(torch.nn.Module):
    """
    Implementation of the mask model for the self-adaptive weights of the
    :class:`SelfAdaptivePINN` solver.
    """

    def __init__(self, func, num_points):
        """
        Initialization of the :class:`Weights` class.

        :param torch.nn.Module func: the mask model.
        :param int num_points: the number of input points.
        """
        super().__init__()

        # Check consistency
        check_consistency(func, torch.nn.Module)

        # Initialize the weights as a learnable parameter
        self.sa_weights = torch.nn.Parameter(torch.zeros(num_points, 1))
        self.func = func

    def forward(self):
        """
        Forward pass implementation for the mask module.

        :return: evaluation of self adaptive weights through the mask.
        :rtype: torch.Tensor
        """
        return self.func(self.sa_weights)


class SelfAdaptivePINN(PINNInterface, MultiSolverInterface):
    r"""
    Self-Adaptive Physics-Informed Neural Network (SelfAdaptivePINN) solver
    class. This class implements the Self-Adaptive Physics-Informed Neural
    Network solver, using a user specified ``model`` to solve a specific
    ``problem``. It can be used to solve both forward and inverse problems.

    The Self-Adapive Physics-Informed Neural Network solver aims to find the
    solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a differential
    problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}
    
    integrating pointwise loss evaluation using a mask :math:m and self-adaptive
    weights, which allow the model to focus on regions of the domain where the
    residual is higher.

    The loss function to solve the problem is

    .. math::

        \mathcal{L}_{\rm{problem}} = \frac{1}{N} \sum_{i=1}^{N_\Omega} m
        \left( \lambda_{\Omega}^{i} \right) \mathcal{L} \left( \mathcal{A}
        [\mathbf{u}](\mathbf{x}) \right) + \frac{1}{N} 
        \sum_{i=1}^{N_{\partial\Omega}}
        m \left( \lambda_{\partial\Omega}^{i} \right) \mathcal{L} 
        \left( \mathcal{B}[\mathbf{u}](\mathbf{x})
        \right),
    
    denoting the self adaptive weights as
    :math:`\lambda_{\Omega}^1, \dots, \lambda_{\Omega}^{N_\Omega}` and
    :math:`\lambda_{\partial \Omega}^1, \dots, 
    \lambda_{\Omega}^{N_\partial \Omega}`
    for :math:`\Omega` and :math:`\partial \Omega`, respectively.

    The Self-Adaptive Physics-Informed Neural Network solver identifies the
    solution and appropriate self adaptive weights by solving the following
    optimization problem:

    .. math::

        \min_{w} \max_{\lambda_{\Omega}^k, \lambda_{\partial \Omega}^s}
        \mathcal{L} ,
    
    where :math:`w` denotes the network parameters, and :math:`\mathcal{L}` is a
    specific loss function, , typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::
        **Original reference**: McClenny, Levi D., and Ulisses M. Braga-Neto.
        *Self-adaptive physics-informed neural networks.*
        Journal of Computational Physics 474 (2023): 111722.
        DOI: `10.1016/j.jcp.2022.111722
        <https://doi.org/10.1016/j.jcp.2022.111722>`_.
    """

    def __init__(
        self,
        problem,
        model,
        weight_function=torch.nn.Sigmoid(),
        optimizer_model=None,
        optimizer_weights=None,
        scheduler_model=None,
        scheduler_weights=None,
        weighting=None,
        loss=None,
    ):
        """
        Initialization of the :class:`SelfAdaptivePINN` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The model to be used.
        :param torch.nn.Module weight_function: The Self-Adaptive mask model.
            Default is ``torch.nn.Sigmoid()``.
        :param Optimizer optimizer_model: The optimizer of the ``model``.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Optimizer optimizer_weights: The optimizer of the
            ``weight_function``.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler_model: Learning rate scheduler for the
            ``model``.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param Scheduler scheduler_weights: Learning rate scheduler for the
            ``weight_function``.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        """
        # Check consistency
        check_consistency(weight_function, torch.nn.Module)

        # Define a ModuleDict for the weights
        weights = {}
        for cond, data in problem.input_pts.items():
            weights[cond] = Weights(func=weight_function, num_points=len(data))
        weights = torch.nn.ModuleDict(weights)

        super().__init__(
            models=[model, weights],
            problem=problem,
            optimizers=[optimizer_model, optimizer_weights],
            schedulers=[scheduler_model, scheduler_weights],
            weighting=weighting,
            loss=loss,
        )

        # Extract the reduction method from the loss function
        self._reduction = self._loss_fn.reduction

        # Set the loss function to return non-aggregated losses
        self._loss_fn = type(self._loss_fn)(reduction="none")

    def training_step(self, batch, batch_idx, **kwargs):
        """
        Solver training step. It computes the optimization cycle and aggregates
        the losses using the ``weighting`` attribute.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param int batch_idx: The index of the current batch.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the training step.
        :rtype: torch.Tensor
        """
        # Weights optimization
        self.optimizer_weights.instance.zero_grad()
        loss = self._optimization_cycle(
            batch=batch, batch_idx=batch_idx, **kwargs
        )
        self.manual_backward(-loss)
        self.optimizer_weights.instance.step()
        self.scheduler_weights.instance.step()

        # Model optimization
        self.optimizer_model.instance.zero_grad()
        loss = self._optimization_cycle(
            batch=batch, batch_idx=batch_idx, **kwargs
        )
        self.manual_backward(loss)
        self.optimizer_model.instance.step()
        self.scheduler_model.instance.step()

        # Log the loss
        self.store_log("train_loss", loss, self.get_batch_size(batch))

        return loss

    @torch.set_grad_enabled(True)
    def validation_step(self, batch, **kwargs):
        """
        The validation step for the Self-Adaptive PINN solver. It returns the
        average residual computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the validation step.
        :rtype: torch.Tensor
        """
        losses = self.optimization_cycle(batch=batch, **kwargs)

        # Aggregate losses for each condition
        for cond, loss in losses.items():
            losses[cond] = self._apply_reduction(loss=losses[cond])

        loss = (sum(losses.values()) / len(losses)).as_subclass(torch.Tensor)
        self.store_log("val_loss", loss, self.get_batch_size(batch))
        return loss

    @torch.set_grad_enabled(True)
    def test_step(self, batch, **kwargs):
        """
        The test step for the Self-Adaptive PINN solver. It returns the average
        residual computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the test step.
        :rtype: torch.Tensor
        """
        losses = self.optimization_cycle(batch=batch, **kwargs)

        # Aggregate losses for each condition
        for cond, loss in losses.items():
            losses[cond] = self._apply_reduction(loss=losses[cond])

        loss = (sum(losses.values()) / len(losses)).as_subclass(torch.Tensor)
        self.store_log("test_loss", loss, self.get_batch_size(batch))
        return loss

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        residuals = self.compute_residual(samples, equation)
        return self._loss_fn(residuals, torch.zeros_like(residuals))

    def loss_data(self, input, target):
        """
        Compute the data loss for the Self-Adaptive PINN solver by evaluating
        the loss between the network's output and the true solution. This method
        should not be overridden, if not intentionally.

        :param input: The input to the neural network.
        :type input: LabelTensor | torch.Tensor
        :param target: The target to compare with the network's output.
        :type target: LabelTensor | torch.Tensor
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor | torch.Tensor
        """
        return self._loss_fn(self.forward(input), target)

    def forward(self, x):
        """
        Forward pass.

        :param x: Input tensor.
        :type x: torch.Tensor | LabelTensor
        :return: The output of the neural network.
        :rtype: torch.Tensor | LabelTensor
        """
        return self.model(x)

    def configure_optimizers(self):
        """
        Optimizer configuration.

        :return: The optimizers and the schedulers
        :rtype: tuple[list[Optimizer], list[Scheduler]]
        """
        # Hook the optimizers to the models
        self.optimizer_model.hook(self.model.parameters())
        self.optimizer_weights.hook(self.weights.parameters())

        # Add unknown parameters to optimization list in case of InverseProblem
        if isinstance(self.problem, InverseProblem):
            self.optimizer_model.instance.add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )

        # Hook the schedulers to the optimizers
        self.scheduler_model.hook(self.optimizer_model)
        self.scheduler_weights.hook(self.optimizer_weights)

        return (
            [self.optimizer_model.instance, self.optimizer_weights.instance],
            [self.scheduler_model.instance, self.scheduler_weights.instance],
        )

    def _optimization_cycle(self, batch, batch_idx, **kwargs):
        """
        Aggregate the loss for each condition in the batch.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param int batch_idx: The index of the current batch.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        # Compute non-aggregated residuals
        residuals = self.optimization_cycle(batch)

        # Compute losses
        losses = {}
        for cond, res in residuals.items():

            weight_tensor = self.weights[cond]()

            # Get the correct indices for the weights. Modulus is used according
            # to the number of points in the condition, as in the PinaDataset.
            len_res = len(res)
            idx = torch.arange(
                batch_idx * len_res,
                (batch_idx + 1) * len_res,
                device=res.device,
            ) % len(self.problem.input_pts[cond])

            # Apply the weights to the residuals
            losses[cond] = self._apply_reduction(
                loss=(res * weight_tensor[idx])
            )

            # Store log
            self.store_log(
                f"{cond}_loss", losses[cond].item(), self.get_batch_size(batch)
            )

        # Clamp unknown parameters in InverseProblem (if needed)
        self._clamp_params()

        # Aggregate
        loss = self.weighting.aggregate(losses).as_subclass(torch.Tensor)

        return loss

    def _apply_reduction(self, loss):
        """
        Apply the specified reduction to the loss. The reduction is deferred
        until the end of the optimization cycle to allow self-adaptive weights
        to be applied to each point beforehand.

        :param torch.Tensor loss: The loss tensor to be reduced.
        :return: The reduced loss tensor.
        :rtype: torch.Tensor
        :raises ValueError: If the reduction method is neither "mean" nor "sum".
        """
        # Apply the specified reduction method
        if self._reduction == "mean":
            return loss.mean()
        if self._reduction == "sum":
            return loss.sum()

        # Raise an error if the reduction method is not recognized
        raise ValueError(
            f"Unknown reduction: {self._reduction}."
            " Supported reductions are 'mean' and 'sum'."
        )

    @property
    def model(self):
        """
        The model.

        :return: The model.
        :rtype: torch.nn.Module
        """
        return self.models[0]

    @property
    def weights(self):
        """
        The self-adaptive weights.

        :return: The self-adaptive weights.
        :rtype: torch.nn.Module
        """
        return self.models[1]

    @property
    def scheduler_model(self):
        """
        The scheduler associated to the model.

        :return: The scheduler for the model.
        :rtype: Scheduler
        """
        return self.schedulers[0]

    @property
    def scheduler_weights(self):
        """
        The scheduler associated to the mask model.

        :return: The scheduler for the mask model.
        :rtype: Scheduler
        """
        return self.schedulers[1]

    @property
    def optimizer_model(self):
        """
        Returns the optimizer associated to the model.

        :return: The optimizer for the model.
        :rtype: Optimizer
        """
        return self.optimizers[0]

    @property
    def optimizer_weights(self):
        """
        The optimizer associated to the mask model.

        :return: The optimizer for the mask model.
        :rtype: Optimizer
        """
        return self.optimizers[1]



"""Module for the Supervised solver."""

from .supervised_solver_interface import SupervisedSolverInterface
from ..solver import SingleSolverInterface


class SupervisedSolver(SupervisedSolverInterface, SingleSolverInterface):
    r"""
    Supervised Solver solver class. This class implements a Supervised Solver,
    using a user specified ``model`` to solve a specific ``problem``.

    The  Supervised Solver class aims to find a map between the input
    :math:`\mathbf{s}:\Omega\rightarrow\mathbb{R}^m` and the output
    :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`.

    Given a model :math:`\mathcal{M}`, the following loss function is
    minimized during training:

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathbf{u}_i - \mathcal{M}(\mathbf{s}_i)),

    where :math:`\mathcal{L}` is a specific loss function, typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    In this context, :math:`\mathbf{u}_i` and :math:`\mathbf{s}_i` indicates
    the will to approximate multiple (discretised) functions given multiple
    (discretised) input functions.
    """

    def __init__(
        self,
        problem,
        model,
        loss=None,
        optimizer=None,
        scheduler=None,
        weighting=None,
        use_lt=True,
    ):
        """
        Initialization of the :class:`SupervisedSolver` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The neural network model to be used.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param bool use_lt: If ``True``, the solver uses LabelTensors as input.
            Default is ``True``.
        """
        super().__init__(
            model=model,
            problem=problem,
            loss=loss,
            optimizer=optimizer,
            scheduler=scheduler,
            weighting=weighting,
            use_lt=use_lt,
        )

    def loss_data(self, input, target):
        """
        Compute the data loss for the Supervised solver by evaluating the loss
        between the network's output and the true solution. This method should
        not be overridden, if not intentionally.

        :param input: The input to the neural network.
        :type input: LabelTensor | torch.Tensor | Graph | Data
        :param target: The target to compare with the network's output.
        :type target: LabelTensor | torch.Tensor | Graph | Data
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor | torch.Tensor | Graph | Data
        """
        return self._loss_fn(self.forward(input), target)



"""Formulation of the Allen Cahn problem."""

import torch
from ... import Condition
from ...problem import SpatialProblem, TimeDependentProblem
from ...equation import Equation, AllenCahn
from ...utils import check_consistency
from ...domain import CartesianDomain


def initial_condition(input_, output_):
    """
    Definition of the initial condition of the Allen Cahn problem.

    :param LabelTensor input_: The input data of the problem.
    :param LabelTensor output_: The output data of the problem.
    :return: The residual of the initial condition.
    :rtype: LabelTensor
    """
    x = input_.extract("x")
    u_0 = x**2 * torch.cos(torch.pi * x)
    return output_ - u_0


class AllenCahnProblem(TimeDependentProblem, SpatialProblem):
    r"""
    Implementation of the Allen Cahn problem in the spatial interval
    :math:`[-1, 1]` and temporal interval :math:`[0, 1]`.

    .. seealso::
        **Original reference**: Sokratis J. Anagnostopoulos, Juan D. Toscano,
        Nikolaos Stergiopulos, and George E. Karniadakis.
        *Residual-based attention and connection to information
        bottleneck theory in PINNs*.
        Computer Methods in Applied Mechanics and Engineering 421 (2024): 116805
        DOI: `10.1016/
        j.cma.2024.116805 <https://doi.org/10.1016/j.cma.2024.116805>`_.

    :Example:
        >>> problem = AllenCahnProblem()
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [-1, 1]})
    temporal_domain = CartesianDomain({"t": [0, 1]})

    domains = {
        "D": CartesianDomain({"x": [-1, 1], "t": [0, 1]}),
        "t0": CartesianDomain({"x": [-1, 1], "t": 0.0}),
    }

    conditions = {
        "t0": Condition(domain="t0", equation=Equation(initial_condition)),
    }

    def __init__(self, alpha=1e-4, beta=5):
        """
        Initialization of the :class:`AllenCahnProblem`.

        :param alpha: The diffusion coefficient.
        :type alpha: float | int
        :param beta: The reaction coefficient.
        :type beta: float | int
        """
        super().__init__()
        check_consistency(alpha, (float, int))
        check_consistency(beta, (float, int))
        self.alpha = alpha
        self.beta = beta

        self.conditions["D"] = Condition(
            domain="D",
            equation=AllenCahn(alpha=self.alpha, beta=self.beta),
        )



"""Module for the Normalizer callback."""

import torch
from lightning.pytorch import Callback
from ..label_tensor import LabelTensor
from ..utils import check_consistency, is_function
from ..condition import InputTargetCondition
from ..data.dataset import PinaGraphDataset


class NormalizerDataCallback(Callback):
    r"""
    A Callback used to normalize the dataset inputs or targets according to
    user-provided scale and shift functions.

    The transformation is applied as:

    .. math::

        x_{\text{new}} = \frac{x - \text{shift}}{\text{scale}}

    :Example:

    >>> NormalizerDataCallback()
    >>> NormalizerDataCallback(
    ...     scale_fn: torch.std,
    ...     shift_fn: torch.mean,
    ...     stage: "all",
    ...     apply_to: "input",
    ... )
    """

    def __init__(
        self,
        scale_fn=torch.std,
        shift_fn=torch.mean,
        stage="all",
        apply_to="input",
    ):
        """
        Initialization of the :class:`NormalizerDataCallback` class.

        :param Callable scale_fn: The function to compute the scaling factor.
            Default is  ``torch.std``.
        :param Callable shift_fn: The function to compute the shifting factor.
            Default is ``torch.mean``.
        :param str stage: The stage in which normalization is applied.
            Accepted values are "train", "validate", "test", or "all".
            Default is ``"all"``.
        :param str apply_to: Whether to normalize "input" or "target" data.
            Default is ``"input"``.
        :raises ValueError: If ``scale_fn`` is not callable.
        :raises ValueError: If ``shift_fn`` is not callable.
        """
        super().__init__()

        # Validate parameters
        self.apply_to = self._validate_apply_to(apply_to)
        self.stage = self._validate_stage(stage)

        # Validate functions
        if not is_function(scale_fn):
            raise ValueError(f"scale_fn must be Callable, got {scale_fn}")
        if not is_function(shift_fn):
            raise ValueError(f"shift_fn must be Callable, got {shift_fn}")
        self.scale_fn = scale_fn
        self.shift_fn = shift_fn

        # Initialize normalizer dictionary
        self._normalizer = {}

    def _validate_apply_to(self, apply_to):
        """
        Validate the ``apply_to`` parameter.

        :param str apply_to: The candidate value for the ``apply_to`` parameter.
        :raises ValueError: If ``apply_to`` is neither "input" nor "target".
        :return: The validated ``apply_to`` value.
        :rtype: str
        """
        check_consistency(apply_to, str)
        if apply_to not in {"input", "target"}:
            raise ValueError(
                f"apply_to must be either 'input' or 'target', got {apply_to}"
            )

        return apply_to

    def _validate_stage(self, stage):
        """
        Validate the ``stage`` parameter.

        :param str stage: The candidate value for the ``stage`` parameter.
        :raises ValueError: If ``stage`` is not one of "train", "validate",
            "test", or "all".
        :return: The validated ``stage`` value.
        :rtype: str
        """
        check_consistency(stage, str)
        if stage not in {"train", "validate", "test", "all"}:
            raise ValueError(
                "stage must be one of 'train', 'validate', 'test', or 'all',"
                f" got {stage}"
            )

        return stage

    def setup(self, trainer, pl_module, stage):
        """
        Apply normalization during setup.

        :param Trainer trainer: A :class:`~pina.trainer.Trainer` instance.
        :param SolverInterface pl_module: A
            :class:`~pina.solver.solver.SolverInterface` instance.
        :param str stage: The current stage.
        :raises RuntimeError: If the training dataset is not available when
            computing normalization parameters.
        :return: The result of the parent setup.
        :rtype: Any

        :raises NotImplementedError: If the dataset is graph-based.
        """

        # Ensure datsets are not graph-based
        if isinstance(trainer.datamodule.train_dataset, PinaGraphDataset):
            raise NotImplementedError(
                "NormalizerDataCallback is not compatible with "
                "graph-based datasets."
            )

        # Extract conditions
        conditions_to_normalize = [
            name
            for name, cond in pl_module.problem.conditions.items()
            if isinstance(cond, InputTargetCondition)
        ]

        # Compute scale and shift parameters
        if not self.normalizer:
            if not trainer.datamodule.train_dataset:
                raise RuntimeError(
                    "Training dataset is not available. Cannot compute "
                    "normalization parameters."
                )
            self._compute_scale_shift(
                conditions_to_normalize, trainer.datamodule.train_dataset
            )

        # Apply normalization based on the specified stage
        if stage == "fit" and self.stage in ["train", "all"]:
            self.normalize_dataset(trainer.datamodule.train_dataset)
        if stage == "fit" and self.stage in ["validate", "all"]:
            self.normalize_dataset(trainer.datamodule.val_dataset)
        if stage == "test" and self.stage in ["test", "all"]:
            self.normalize_dataset(trainer.datamodule.test_dataset)

        return super().setup(trainer, pl_module, stage)

    def _compute_scale_shift(self, conditions, dataset):
        """
        Compute scale and shift parameters for each condition in the dataset.

        :param list conditions: The list of condition names.
        :param dataset: The `~pina.data.dataset.PinaDataset` dataset.
        """
        for cond in conditions:
            if cond in dataset.conditions_dict:
                data = dataset.conditions_dict[cond][self.apply_to]
                shift = self.shift_fn(data)
                scale = self.scale_fn(data)
                self._normalizer[cond] = {
                    "shift": shift,
                    "scale": scale,
                }

    @staticmethod
    def _norm_fn(value, scale, shift):
        """
        Normalize a value according to the scale and shift parameters.

        :param value: The input tensor to normalize.
        :type value: torch.Tensor | LabelTensor
        :param float scale: The scaling factor.
        :param float shift: The shifting factor.
        :return: The normalized tensor.
        :rtype: torch.Tensor | LabelTensor
        """
        scaled_value = (value - shift) / scale
        if isinstance(value, LabelTensor):
            scaled_value = LabelTensor(scaled_value, value.labels)

        return scaled_value

    def normalize_dataset(self, dataset):
        """
        Apply in-place normalization to the dataset.

        :param PinaDataset dataset: The dataset to be normalized.
        """
        # Initialize update dictionary
        update_dataset_dict = {}

        # Iterate over conditions and apply normalization
        for cond, norm_params in self.normalizer.items():
            points = dataset.conditions_dict[cond][self.apply_to]
            scale = norm_params["scale"]
            shift = norm_params["shift"]
            normalized_points = self._norm_fn(points, scale, shift)
            update_dataset_dict[cond] = {
                self.apply_to: (
                    LabelTensor(normalized_points, points.labels)
                    if isinstance(points, LabelTensor)
                    else normalized_points
                )
            }

        # Update the dataset in-place
        dataset.update_data(update_dataset_dict)

    @property
    def normalizer(self):
        """
        Get the dictionary of normalization parameters.

        :return: The dictionary of normalization parameters.
        :rtype: dict
        """
        return self._normalizer



"""Module for the Kernel Neural Operator model class."""

import torch
from ..utils import check_consistency


class KernelNeuralOperator(torch.nn.Module):
    r"""
    Base class for Neural Operators with integral kernels.

    This class serves as a foundation for building Neural Operators that
    incorporate multiple integral kernels. All Neural Operator models in
    PINA inherit from this class. The design follows the framework proposed
    by Kovachki et al., as illustrated in Figure 2 of their work.

    Neural Operators derived from this class can be expressed as:

    .. math::
        G_\theta  := P \circ K_m \circ \cdot \circ K_1 \circ L

    where:

    *   :math:`G_\theta: \mathcal{A}\subset \mathbb{R}^{\rm{in}} \rightarrow
        \mathcal{D}\subset \mathbb{R}^{\rm{out}}` is the neural operator
        approximation of the unknown real operator :math:`G`, that is
        :math:`G \approx G_\theta`
    *   :math:`L: \mathcal{A}\subset \mathbb{R}^{\rm{in}} \rightarrow
        \mathbb{R}^{\rm{emb}}` is a lifting operator mapping the input
        from its domain :math:`\mathcal{A}\subset \mathbb{R}^{\rm{in}}`
        to its embedding dimension :math:`\mathbb{R}^{\rm{emb}}`
    *   :math:`\{K_i : \mathbb{R}^{\rm{emb}} \rightarrow
        \mathbb{R}^{\rm{emb}} \}_{i=1}^m` are :math:`m` integral kernels
        mapping each hidden representation to the next one.
    *   :math:`P : \mathbb{R}^{\rm{emb}} \rightarrow  \mathcal{D}\subset
        \mathbb{R}^{\rm{out}}` is a projection operator mapping the hidden
        representation to the output function.

    .. seealso::

        **Original reference**: Kovachki, N., Li, Z., Liu, B.,
        Azizzadenesheli, K., Bhattacharya, K., Stuart, A., & Anandkumar, A.
        (2023).
        *Neural operator: Learning maps between function spaces with
        applications to PDEs*.
        Journal of Machine Learning Research, 24(89), 1-97.
    """

    def __init__(self, lifting_operator, integral_kernels, projection_operator):
        """
        Initialization of the :class:`KernelNeuralOperator` class.

        :param torch.nn.Module lifting_operator: The lifting operator mapping
            the input to its hidden dimension.
        :param torch.nn.Module integral_kernels: List of integral kernels
            mapping each hidden representation to the next one.
        :param torch.nn.Module projection_operator: The projection operator
            mapping the hidden representation to the output function.
        """

        super().__init__()

        self._lifting_operator = lifting_operator
        self._integral_kernels = integral_kernels
        self._projection_operator = projection_operator

    @property
    def lifting_operator(self):
        """
        The lifting operator module.

        :return: The lifting operator module.
        :rtype: torch.nn.Module
        """
        return self._lifting_operator

    @lifting_operator.setter
    def lifting_operator(self, value):
        """
        Set the lifting operator module.

        :param torch.nn.Module value: The lifting operator module.
        """
        check_consistency(value, torch.nn.Module)
        self._lifting_operator = value

    @property
    def projection_operator(self):
        """
        The projection operator module.

        :return: The projection operator module.
        :rtype: torch.nn.Module
        """
        return self._projection_operator

    @projection_operator.setter
    def projection_operator(self, value):
        """
        Set the projection operator module.

        :param torch.nn.Module value: The projection operator module.
        """
        check_consistency(value, torch.nn.Module)
        self._projection_operator = value

    @property
    def integral_kernels(self):
        """
        The integral kernels operator module.

        :return: The integral kernels operator module.
        :rtype: torch.nn.Module
        """
        return self._integral_kernels

    @integral_kernels.setter
    def integral_kernels(self, value):
        """
        Set the integral kernels operator module.

        :param torch.nn.Module value: The integral kernels operator module.
        """
        check_consistency(value, torch.nn.Module)
        self._integral_kernels = value

    def forward(self, x):
        r"""
        Forward pass for the :class:`KernelNeuralOperator` model.

        The ``lifting_operator`` maps the input to the hidden dimension.
        The ``integral_kernels`` apply the integral kernels to the hidden
        representation. The ``projection_operator`` maps the hidden
        representation to the output function.

        :param x: The input tensor for performing the computation. It expects
            a tensor :math:`B \times N \times D`, where :math:`B` is the
            batch_size, :math:`N` the number of points in the mesh, and
            :math:`D` the dimension of the problem. In particular, :math:`D`
            is the number of spatial, parametric, and/or temporal variables
            plus the field variables. For instance, for 2D problems with 2
            output variables, :math:`D=4`.
        :type x: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        x = self.lifting_operator(x)
        x = self.integral_kernels(x)
        x = self.projection_operator(x)
        return x



"""Module for the Supervised solvers."""

__all__ = [
    "SupervisedSolverInterface",
    "SupervisedSolver",
    "ReducedOrderModelSolver",
]

from .supervised_solver_interface import SupervisedSolverInterface
from .supervised import SupervisedSolver
from .reduced_order_model import ReducedOrderModelSolver



"""Module for the Radial Basis Function Interpolation layer."""

import math
import warnings
from itertools import combinations_with_replacement
import torch
from ...utils import check_consistency


def linear(r):
    """
    Linear radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The linear radial basis function.
    :rtype: torch.Tensor
    """
    return -r


def thin_plate_spline(r, eps=1e-7):
    """
    Thin plate spline radial basis function.

    :param torch.Tensor r: Distance between points.
    :param float eps: Small value to avoid log(0).
    :return: The thin plate spline radial basis function.
    :rtype: torch.Tensor
    """
    r = torch.clamp(r, min=eps)
    return r**2 * torch.log(r)


def cubic(r):
    """
    Cubic radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The cubic radial basis function.
    :rtype: torch.Tensor
    """
    return r**3


def quintic(r):
    """
    Quintic radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The quintic radial basis function.
    :rtype: torch.Tensor
    """
    return -(r**5)


def multiquadric(r):
    """
    Multiquadric radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The multiquadric radial basis function.
    :rtype: torch.Tensor
    """
    return -torch.sqrt(r**2 + 1)


def inverse_multiquadric(r):
    """
    Inverse multiquadric radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The inverse multiquadric radial basis function.
    :rtype: torch.Tensor
    """
    return 1 / torch.sqrt(r**2 + 1)


def inverse_quadratic(r):
    """
    Inverse quadratic radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The inverse quadratic radial basis function.
    :rtype: torch.Tensor
    """
    return 1 / (r**2 + 1)


def gaussian(r):
    """
    Gaussian radial basis function.

    :param torch.Tensor r: Distance between points.
    :return: The gaussian radial basis function.
    :rtype: torch.Tensor
    """
    return torch.exp(-(r**2))


radial_functions = {
    "linear": linear,
    "thin_plate_spline": thin_plate_spline,
    "cubic": cubic,
    "quintic": quintic,
    "multiquadric": multiquadric,
    "inverse_multiquadric": inverse_multiquadric,
    "inverse_quadratic": inverse_quadratic,
    "gaussian": gaussian,
}

scale_invariant = {"linear", "thin_plate_spline", "cubic", "quintic"}

min_degree_funcs = {
    "multiquadric": 0,
    "linear": 0,
    "thin_plate_spline": 1,
    "cubic": 1,
    "quintic": 2,
}


class RBFBlock(torch.nn.Module):
    """
    Radial Basis Function (RBF) interpolation layer.

    The user needs to fit the model with the data, before using it to
    interpolate new points. The layer is not trainable.

    .. note::
        It reproduces the implementation of :class:`scipy.interpolate.RBFBlock`
        and it is inspired from the implementation in `torchrbf.
        <https://github.com/ArmanMaesumi/torchrbf>`_
    """

    def __init__(
        self,
        neighbors=None,
        smoothing=0.0,
        kernel="thin_plate_spline",
        epsilon=None,
        degree=None,
    ):
        """
        Initialization of the :class:`RBFBlock` class.

        :param int neighbors: The number of neighbors used for interpolation.
            If ``None``, all data are used.
        :param float smoothing: The moothing parameter for the interpolation.
            If ``0.0``, the interpolation is exact and no smoothing is applied.
        :param str kernel: The radial basis function to use.
            The available kernels are: ``linear``, ``thin_plate_spline``,
            ``cubic``, ``quintic``, ``multiquadric``, ``inverse_multiquadric``,
            ``inverse_quadratic``, or ``gaussian``.
        :param float epsilon: The shape parameter that scales the input to the
            RBF. Default is ``1`` for kernels in the ``scale_invariant``
            dictionary, while it must be specified for other kernels.
        :param int degree: The degree of the polynomial. Some kernels require a
            minimum degree of the polynomial to ensure that the RBF is well
            defined. These minimum degrees are specified in the
            ``min_degree_funcs`` dictionary. If ``degree`` is less than the
            minimum degree required, a warning is raised and the degree is set
            to the minimum value.
        """

        super().__init__()
        check_consistency(neighbors, (int, type(None)))
        check_consistency(smoothing, (int, float, torch.Tensor))
        check_consistency(kernel, str)
        check_consistency(epsilon, (float, type(None)))
        check_consistency(degree, (int, type(None)))

        self.neighbors = neighbors
        self.smoothing = smoothing
        self.kernel = kernel
        self.epsilon = epsilon
        self.degree = degree
        self.powers = None
        # initialize data points and values
        self.y = None
        self.d = None
        # initialize attributes for the fitted model
        self._shift = None
        self._scale = None
        self._coeffs = None

    @property
    def smoothing(self):
        """
        The smoothing parameter for the interpolation.

        :return: The smoothing parameter.
        :rtype: float
        """
        return self._smoothing

    @smoothing.setter
    def smoothing(self, value):
        """
        Set the smoothing parameter for the interpolation.

        :param float value: The smoothing parameter.
        """
        self._smoothing = value

    @property
    def kernel(self):
        """
        The Radial basis function.

        :return: The radial basis function.
        :rtype: str
        """
        return self._kernel

    @kernel.setter
    def kernel(self, value):
        """
        Set the radial basis function.

        :param str value: The radial basis function.
        """
        if value not in radial_functions:
            raise ValueError(f"Unknown kernel: {value}")
        self._kernel = value.lower()

    @property
    def epsilon(self):
        """
        The shape parameter that scales the input to the RBF.

        :return: The shape parameter.
        :rtype: float
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):
        """
        Set the shape parameter.

        :param float value: The shape parameter.
        :raises ValueError: If the kernel requires an epsilon and it is not
            specified.
        """
        if value is None:
            if self.kernel in scale_invariant:
                value = 1.0
            else:
                raise ValueError("Must specify `epsilon` for this kernel.")
        else:
            value = float(value)
        self._epsilon = value

    @property
    def degree(self):
        """
        The degree of the polynomial.

        :return: The degree of the polynomial.
        :rtype: int
        """
        return self._degree

    @degree.setter
    def degree(self, value):
        """
        Set the degree of the polynomial.

        :param int value: The degree of the polynomial.
        :raises UserWarning: If the degree is less than the minimum required
            for the kernel.
        :raises ValueError: If the degree is less than -1.
        """
        min_degree = min_degree_funcs.get(self.kernel, -1)
        if value is None:
            value = max(min_degree, 0)
        else:
            value = int(value)
            if value < -1:
                raise ValueError("`degree` must be at least -1.")
            if value < min_degree:
                warnings.warn(
                    "`degree` is too small for this kernel. Setting to "
                    f"{min_degree}.",
                    UserWarning,
                )
        self._degree = value

    def _check_data(self, y, d):
        """
        Check the data consistency.

        :param torch.Tensor y: The tensor of data points.
        :param torch.Tensor d: The tensor of data values.
        :raises ValueError: If the data is not consistent.
        """
        if y.ndim != 2:
            raise ValueError("y must be a 2-dimensional tensor.")

        if d.shape[0] != y.shape[0]:
            raise ValueError(
                "The first dim of d must have the same length as "
                "the first dim of y."
            )

        if isinstance(self.smoothing, (int, float)):
            self.smoothing = (
                torch.full((y.shape[0],), self.smoothing).float().to(y.device)
            )

    def fit(self, y, d):
        """
        Fit the RBF interpolator to the data.

        :param torch.Tensor y: The tensor of data points.
        :param torch.Tensor d: The tensor of data values.
        :raises NotImplementedError: If the neighbors are not ``None``.
        :raises ValueError: If the data is not compatible with the requested
            degree.
        """
        self._check_data(y, d)

        self.y = y
        self.d = d

        if self.neighbors is None:
            nobs = self.y.shape[0]
        else:
            raise NotImplementedError("Neighbors currently not supported")

        powers = RBFBlock.monomial_powers(self.y.shape[1], self.degree).to(
            y.device
        )
        if powers.shape[0] > nobs:
            raise ValueError(
                "The data is not compatible with the requested degree."
            )

        if self.neighbors is None:
            self._shift, self._scale, self._coeffs = RBFBlock.solve(
                self.y,
                self.d.reshape((self.y.shape[0], -1)),
                self.smoothing,
                self.kernel,
                self.epsilon,
                powers,
            )

        self.powers = powers

    def forward(self, x):
        """
        Forward pass.

        :param torch.Tensor x: The tensor of points to interpolate.
        :raises ValueError: If the input is not a 2-dimensional tensor.
        :raises ValueError: If the second dimension of the input is not the same
            as the second dimension of the data.
        :return: The interpolated data.
        :rtype: torch.Tensor
        """
        if x.ndim != 2:
            raise ValueError("`x` must be a 2-dimensional tensor.")

        nx, ndim = x.shape
        if ndim != self.y.shape[1]:
            raise ValueError(
                "Expected the second dim of `x` to have length "
                f"{self.y.shape[1]}."
            )

        kernel_func = radial_functions[self.kernel]

        yeps = self.y * self.epsilon
        xeps = x * self.epsilon
        xhat = (x - self._shift) / self._scale

        kv = RBFBlock.kernel_vector(xeps, yeps, kernel_func)
        p = RBFBlock.polynomial_matrix(xhat, self.powers)
        vec = torch.cat([kv, p], dim=1)
        out = torch.matmul(vec, self._coeffs)
        out = out.reshape((nx,) + self.d.shape[1:])
        return out

    @staticmethod
    def kernel_vector(x, y, kernel_func):
        """
        Evaluate for all points ``x`` the radial functions with center ``y``.

        :param torch.Tensor x: The tensor of points.
        :param torch.Tensor y: The tensor of centers.
        :param str kernel_func: Radial basis function to use.
        :return: The radial function values.
        :rtype: torch.Tensor
        """
        return kernel_func(torch.cdist(x, y))

    @staticmethod
    def polynomial_matrix(x, powers):
        """
        Evaluate monomials of power ``powers`` at points ``x``.

        :param torch.Tensor x: The tensor of points.
        :param torch.Tensor powers: The tensor of powers for each monomial.
        :return: The monomial values.
        :rtype: torch.Tensor
        """
        x_ = torch.repeat_interleave(x, repeats=powers.shape[0], dim=0)
        powers_ = powers.repeat(x.shape[0], 1)
        return torch.prod(x_**powers_, dim=1).view(x.shape[0], powers.shape[0])

    @staticmethod
    def kernel_matrix(x, kernel_func):
        """
        Return the radial function values for all pairs of points in ``x``.

        :param torch.Tensor x: The tensor of points.
        :param str kernel_func: The radial basis function to use.
        :return: The radial function values.
        :rtype: torch.Tensor
        """
        return kernel_func(torch.cdist(x, x))

    @staticmethod
    def monomial_powers(ndim, degree):
        """
        Return the powers for each monomial in a polynomial.

        :param int ndim: The number of variables in the polynomial.
        :param int degree: The degree of the polynomial.
        :return: The powers for each monomial.
        :rtype: torch.Tensor
        """
        nmonos = math.comb(degree + ndim, ndim)
        out = torch.zeros((nmonos, ndim), dtype=torch.int32)
        count = 0
        for deg in range(degree + 1):
            for mono in combinations_with_replacement(range(ndim), deg):
                for var in mono:
                    out[count, var] += 1
                count += 1
        return out

    @staticmethod
    def build(y, d, smoothing, kernel, epsilon, powers):
        """
        Build the RBF linear system.

        :param torch.Tensor y: The tensor of data points.
        :param torch.Tensor d: The tensor of data values.
        :param torch.Tensor smoothing: The tensor of smoothing parameters.
        :param str kernel: The radial basis function to use.
        :param float epsilon: The shape parameter that scales the input to the
            RBF.
        :param torch.Tensor powers: The tensor of powers for each monomial.
        :return: The left-hand side and right-hand side of the linear system,
            and the shift and scale parameters.
        :rtype: tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
        """
        p = d.shape[0]
        s = d.shape[1]
        r = powers.shape[0]
        kernel_func = radial_functions[kernel]

        mins = torch.min(y, dim=0).values
        maxs = torch.max(y, dim=0).values
        shift = (maxs + mins) / 2
        scale = (maxs - mins) / 2

        scale[scale == 0.0] = 1.0

        yeps = y * epsilon
        yhat = (y - shift) / scale

        lhs = torch.empty((p + r, p + r), device=d.device).float()
        lhs[:p, :p] = RBFBlock.kernel_matrix(yeps, kernel_func)
        lhs[:p, p:] = RBFBlock.polynomial_matrix(yhat, powers)
        lhs[p:, :p] = lhs[:p, p:].T
        lhs[p:, p:] = 0.0
        lhs[:p, :p] += torch.diag(smoothing)

        rhs = torch.empty((r + p, s), device=d.device).float()
        rhs[:p] = d
        rhs[p:] = 0.0
        return lhs, rhs, shift, scale

    @staticmethod
    def solve(y, d, smoothing, kernel, epsilon, powers):
        """
        Build and solve the RBF linear system.

        :param torch.Tensor y: The tensor of data points.
        :param torch.Tensor d: The tensor of data values.
        :param torch.Tensor smoothing: The tensor of smoothing parameters.

        :param str kernel: The radial basis function to use.
        :param float epsilon: The shape parameter that scaled the input to the
            RBF.
        :param torch.Tensor powers: The tensor of powers for each monomial.
        :raises ValueError: If the linear system is singular.
        :return: The shift and scale parameters, and the coefficients of the
            interpolator.
        :rtype: tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        """

        lhs, rhs, shift, scale = RBFBlock.build(
            y, d, smoothing, kernel, epsilon, powers
        )
        try:
            coeffs = torch.linalg.solve(lhs, rhs)
        except RuntimeError as e:
            msg = "Singular matrix."
            nmonos = powers.shape[0]
            if nmonos > 0:
                pmat = RBFBlock.polynomial_matrix((y - shift) / scale, powers)
                rank = torch.linalg.matrix_rank(pmat)
                if rank < nmonos:
                    msg = (
                        "Singular matrix. The matrix of monomials evaluated at "
                        "the data point coordinates does not have full column "
                        f"rank ({rank}/{nmonos})."
                    )

            raise ValueError(msg) from e

        return shift, scale, coeffs



"""Formulation of the Helmholtz problem."""

import torch
from ... import Condition
from ...equation import FixedValue, Helmholtz
from ...utils import check_consistency
from ...domain import CartesianDomain
from ...problem import SpatialProblem


class HelmholtzProblem(SpatialProblem):
    r"""
    Implementation of the Helmholtz problem in the square domain
    :math:`[-1, 1] \times [-1, 1]`.

    .. seealso::
        **Original reference**: Si, Chenhao, et al. *Complex Physics-Informed
        Neural Network.* arXiv preprint arXiv:2502.04917 (2025).
        DOI: `arXiv:2502.04917 <https://arxiv.org/abs/2502.04917>`_.

    :Example:
        >>> problem = HelmholtzProblem()
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [-1, 1], "y": [-1, 1]})

    domains = {
        "D": CartesianDomain({"x": [-1, 1], "y": [-1, 1]}),
        "g1": CartesianDomain({"x": [-1, 1], "y": 1.0}),
        "g2": CartesianDomain({"x": [-1, 1], "y": -1.0}),
        "g3": CartesianDomain({"x": 1.0, "y": [-1, 1]}),
        "g4": CartesianDomain({"x": -1.0, "y": [-1, 1]}),
    }

    conditions = {
        "g1": Condition(domain="g1", equation=FixedValue(0.0)),
        "g2": Condition(domain="g2", equation=FixedValue(0.0)),
        "g3": Condition(domain="g3", equation=FixedValue(0.0)),
        "g4": Condition(domain="g4", equation=FixedValue(0.0)),
    }

    def __init__(self, alpha=3.0):
        """
        Initialization of the :class:`HelmholtzProblem` class.

        :param alpha: Parameter of the forcing term.
        :type alpha: float | int
        """
        super().__init__()
        check_consistency(alpha, (int, float))
        self.alpha = alpha

        def forcing_term(input_):
            """
            Implementation of the forcing term.
            """
            return (
                (1 - 2 * (self.alpha * torch.pi) ** 2)
                * torch.sin(self.alpha * torch.pi * input_.extract("x"))
                * torch.sin(self.alpha * torch.pi * input_.extract("y"))
            )

        self.conditions["D"] = Condition(
            domain="D",
            equation=Helmholtz(self.alpha, forcing_term),
        )

    def solution(self, pts):
        """
        Implementation of the analytical solution of the Helmholtz problem.

        :param LabelTensor pts: Points where the solution is evaluated.
        :return: The analytical solution of the Poisson problem.
        :rtype: LabelTensor
        """
        sol = torch.sin(self.alpha * torch.pi * pts.extract("x")) * torch.sin(
            self.alpha * torch.pi * pts.extract("y")
        )
        sol.labels = self.output_variables
        return sol



"""Module for the SwitchOptimizer callback."""

from lightning.pytorch.callbacks import Callback
from ..optim import TorchOptimizer
from ..utils import check_consistency


class SwitchOptimizer(Callback):
    """
    PINA Implementation of a Lightning Callback to switch optimizer during
    training.
    """

    def __init__(self, new_optimizers, epoch_switch):
        """
        This callback allows switching between different optimizers during
        training, enabling the exploration of multiple optimization strategies
        without interrupting the training process.

        :param new_optimizers: The model optimizers to switch to. Can be a
            single :class:`torch.optim.Optimizer` instance or a list of them
            for multiple model solver.
        :type new_optimizers: pina.optim.TorchOptimizer | list
        :param int epoch_switch: The epoch at which the optimizer switch occurs.

        Example:
            >>> optimizer = TorchOptimizer(torch.optim.Adam, lr=0.01)
            >>> switch_callback = SwitchOptimizer(
            >>>     new_optimizers=optimizer, epoch_switch=10
            >>> )
        """
        super().__init__()

        # Check if epoch_switch is greater than 1
        if epoch_switch < 1:
            raise ValueError("epoch_switch must be greater than one.")

        # If new_optimizers is not a list, convert it to a list
        if not isinstance(new_optimizers, list):
            new_optimizers = [new_optimizers]

        # Check consistency
        check_consistency(epoch_switch, int)
        for optimizer in new_optimizers:
            check_consistency(optimizer, TorchOptimizer)

        # Store the new optimizers and epoch switch
        self._new_optimizers = new_optimizers
        self._epoch_switch = epoch_switch

    def on_train_epoch_start(self, trainer, __):
        """
        Switch the optimizer at the start of the specified training epoch.

        :param lightning.pytorch.Trainer trainer: The trainer object managing
            the training process.
        :param _: Placeholder argument (not used).
        """
        # Check if the current epoch matches the switch epoch
        if trainer.current_epoch == self._epoch_switch:
            optims = []

            # Hook the new optimizers to the model parameters
            for idx, optim in enumerate(self._new_optimizers):
                optim.hook(trainer.solver._pina_models[idx].parameters())
                optims.append(optim)

            # Update the solver's optimizers
            trainer.solver._pina_optimizers = optims

            # Update the trainer's strategy optimizers
            trainer.strategy.optimizers = [o.instance for o in optims]



"""Module for the Cartesian Domain."""

import torch

from .domain_interface import DomainInterface
from ..label_tensor import LabelTensor
from ..utils import torch_lhs, chebyshev_roots


class CartesianDomain(DomainInterface):
    """
    Implementation of the hypercube domain.
    """

    def __init__(self, cartesian_dict):
        """
        Initialization of the :class:`CartesianDomain` class.

        :param dict cartesian_dict: A dictionary where the keys are the
            variable names and the values are the domain extrema. The domain
            extrema can be either a list with two elements or a single number.
            If the domain extrema is a single number, the variable is fixed to
            that value.

        :Example:
            >>> spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
        """
        self.fixed_ = {}
        self.range_ = {}

        for k, v in cartesian_dict.items():
            if isinstance(v, (int, float)):
                self.fixed_[k] = v
            elif isinstance(v, (list, tuple)) and len(v) == 2:
                self.range_[k] = v
            else:
                raise TypeError

    @property
    def sample_modes(self):
        """
        List of available sampling modes.

        :return: List of available sampling modes.
        :rtype: list[str]
        """
        return ["random", "grid", "lh", "chebyshev", "latin"]

    @property
    def variables(self):
        """
        List of variables of the domain.

        :return: List of variables of the domain.
        :rtype: list[str]
        """
        return sorted(list(self.fixed_.keys()) + list(self.range_.keys()))

    def update(self, new_domain):
        """
        Add new dimensions to an existing :class:`CartesianDomain` object.

        :param CartesianDomain new_domain: New domain to be added to an existing
            :class:`CartesianDomain` object.

        :Example:
            >>> spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
            >>> spatial_domain.variables
            ['x', 'y']
            >>> spatial_domain_2 = CartesianDomain({'z': [3, 4], 'w': [0, 1]})
            >>> spatial_domain.update(spatial_domain_2)
            >>> spatial_domain.variables
            ['x', 'y', 'z', 'w']
        """
        self.fixed_.update(new_domain.fixed_)
        self.range_.update(new_domain.range_)

    def _sample_range(self, n, mode, bounds):
        """
        Rescale the samples to fit within the specified bounds.

        :param int n: Number of points to sample.
        :param str mode: Sampling method. Default is ``random``.
        :param torch.Tensor bounds: Bounds of the domain.
        :raises RuntimeError: Wrong bounds initialization.
        :raises ValueError: Invalid sampling mode.
        :return: Rescaled sample points.
        :rtype: torch.Tensor
        """
        dim = bounds.shape[0]
        if mode in ["chebyshev", "grid"] and dim != 1:
            raise RuntimeError("Wrong bounds initialization")

        if mode == "random":
            pts = torch.rand(size=(n, dim))
        elif mode == "chebyshev":
            pts = chebyshev_roots(n).mul(0.5).add(0.5).reshape(-1, 1)
        elif mode == "grid":
            pts = torch.linspace(0, 1, n).reshape(-1, 1)
        elif mode in ["lh", "latin"]:
            pts = torch_lhs(n, dim)
        else:
            raise ValueError("Invalid mode")

        return pts * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: Number of points to sample, see Note below for reference.
        :param str mode: Sampling method. Default is ``random``.
            Available modes: random sampling, ``random``;
            latin hypercube sampling, ``latin`` or ``lh``;
            chebyshev sampling, ``chebyshev``; grid sampling ``grid``.
        :param list[str] variables: variables to be sampled. Default is ``all``.
        :return: Sampled points.
        :rtype: LabelTensor

        .. note::
            When multiple variables are involved, the total number of sampled
            points may differ from ``n``, depending on the chosen ``mode``.
            If ``mode`` is ``grid`` or ``chebyshev``, points are sampled
            independently for each variable and then combined, resulting in a
            total number of points equal to ``n`` raised to the power of the
            number of variables. If 'mode' is 'random', ``lh`` or ``latin``,
            all variables are sampled together, and the total number of points
            remains ``n``.

        .. warning::
            The extrema of CartesianDomain are only sampled when using the
            ``grid`` mode.

        :Example:
            >>> spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
            >>> spatial_domain.sample(n=4, mode='random')
                tensor([[0.0108, 0.7643],
                        [0.4477, 0.8015],
                        [0.2063, 0.8087],
                        [0.8735, 0.6349]])
            >>> spatial_domain.sample(n=4, mode='grid')
                tensor([[0.0000, 0.0000],
                        [0.3333, 0.0000],
                        [0.6667, 0.0000],
                        [1.0000, 0.0000],
                        [0.0000, 0.3333],
                        [0.3333, 0.3333],
                        [0.6667, 0.3333],
                        [1.0000, 0.3333],
                        [0.0000, 0.6667],
                        [0.3333, 0.6667],
                        [0.6667, 0.6667],
                        [1.0000, 0.6667],
                        [0.0000, 1.0000],
                        [0.3333, 1.0000],
                        [0.6667, 1.0000],
                        [1.0000, 1.0000]])
        """

        def _1d_sampler(n, mode, variables):
            """
            Sample each variable independently.

            :param int n: Number of points to sample.
            :param str mode: Sampling method.
            :param list[str] variables: variables to be sampled.
            :return: Sampled points.
            :rtype: list[LabelTensor]
            """
            tmp = []
            for variable in variables:
                if variable in self.range_:
                    bound = torch.tensor([self.range_[variable]])
                    pts_variable = self._sample_range(n, mode, bound)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    tmp.append(pts_variable)
            if tmp:
                result = tmp[0]
                for i in tmp[1:]:
                    result = result.append(i, mode="cross")

            for variable in variables:
                if variable in self.fixed_:
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(
                        result.shape[0], 1
                    )
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    result = result.append(pts_variable, mode="std")

            return result

        def _Nd_sampler(n, mode, variables):
            """
            Sample all variables together.

            :param int n: Number of points to sample.
            :param str mode: Sampling method.
            :param list[str] variables: variables to be sampled.
            :return: Sampled points.
            :rtype: list[LabelTensor]
            """
            pairs = [(k, v) for k, v in self.range_.items() if k in variables]
            keys, values = map(list, zip(*pairs))
            bounds = torch.tensor(values)
            result = self._sample_range(n, mode, bounds)
            result = result.as_subclass(LabelTensor)
            result.labels = keys

            for variable in variables:
                if variable in self.fixed_:
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(
                        result.shape[0], 1
                    )
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    result = result.append(pts_variable, mode="std")
            return result

        def _single_points_sample(n, variables):
            """
            Sample a single point in one dimension.

            :param int n: Number of points to sample.
            :param list[str] variables: variables to be sampled.
            :return: Sampled points.
            :rtype: list[torch.Tensor]
            """
            tmp = []
            for variable in variables:
                if variable in self.fixed_:
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(n, 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]
                    tmp.append(pts_variable)

            result = tmp[0]
            for i in tmp[1:]:
                result = result.append(i, mode="std")

            return result

        if variables == "all":
            variables = self.variables
        elif isinstance(variables, (list, tuple)):
            variables = sorted(variables)

        if self.fixed_ and (not self.range_):
            return _single_points_sample(n, variables)
        if isinstance(variables, str) and variables in self.fixed_:
            return _single_points_sample(n, variables)

        if mode in ["grid", "chebyshev"]:
            return _1d_sampler(n, mode, variables).extract(variables)
        if mode in ["random", "lh", "latin"]:
            return _Nd_sampler(n, mode, variables).extract(variables)
        raise ValueError(f"mode={mode} is not valid.")

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the hypercube.

        :param LabelTensor point: Point to be checked.
        :param bool check_border: If ``True``, the border is considered inside
            the hypercube. Default is ``False``.
        :return: ``True`` if the point is inside the domain,
            ``False`` otherwise.
        :rtype: bool
        """
        is_inside = []

        # check fixed variables
        for variable, value in self.fixed_.items():
            if variable in point.labels:
                is_inside.append(point.extract([variable]) == value)

        # check not fixed variables
        for variable, bound in self.range_.items():
            if variable in point.labels:

                if check_border:
                    check = bound[0] <= point.extract([variable]) <= bound[1]
                else:
                    check = bound[0] < point.extract([variable]) < bound[1]

                is_inside.append(check)

        return all(is_inside)



"""Module for the Competitive PINN solver."""

import copy
import torch

from ...problem import InverseProblem
from .pinn_interface import PINNInterface
from ..solver import MultiSolverInterface


class CompetitivePINN(PINNInterface, MultiSolverInterface):
    r"""
    Competitive Physics-Informed Neural Network (CompetitivePINN) solver class.
    This class implements the Competitive Physics-Informed Neural Network
    solver, using a user specified ``model`` to solve a specific ``problem``.
    It can be used to solve both forward and inverse problems.

    The Competitive Physics-Informed Neural Network solver aims to find the
    solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a differential
    problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function with respect to the model parameters, while 
    maximizing it with respect to the discriminator parameters:

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(D(\mathbf{x}_i)\mathcal{A}[\mathbf{u}](\mathbf{x}_i))+
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(D(\mathbf{x}_i)\mathcal{B}[\mathbf{u}](\mathbf{x}_i)),

    where :math:D is the discriminator network, which identifies the points
    where the model performs worst, and :math:\mathcal{L} is a specific loss
    function, typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Zeng, Qi, et al.
        *Competitive physics informed networks.*
        International Conference on Learning Representations, ICLR 2022
        `OpenReview Preprint <https://openreview.net/forum?id=z9SIj-IM7tn>`_.
    """

    def __init__(
        self,
        problem,
        model,
        discriminator=None,
        optimizer_model=None,
        optimizer_discriminator=None,
        scheduler_model=None,
        scheduler_discriminator=None,
        weighting=None,
        loss=None,
    ):
        """
        Initialization of the :class:`CompetitivePINN` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The neural network model to be used.
        :param torch.nn.Module discriminator: The discriminator to be used.
            If ``None``, the discriminator is a deepcopy of the ``model``.
            Default is ``None``.
        :param torch.optim.Optimizer optimizer_model: The optimizer of the
            ``model``. If ``None``, the :class:`torch.optim.Adam` optimizer is
            used. Default is ``None``.
        :param torch.optim.Optimizer optimizer_discriminator: The optimizer of
            the ``discriminator``. If ``None``, the :class:`torch.optim.Adam`
            optimizer is used. Default is ``None``.
        :param Scheduler scheduler_model: Learning rate scheduler for the
            ``model``.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param Scheduler scheduler_discriminator: Learning rate scheduler for
            the ``discriminator``.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        """
        if discriminator is None:
            discriminator = copy.deepcopy(model)

        super().__init__(
            models=[model, discriminator],
            problem=problem,
            optimizers=[optimizer_model, optimizer_discriminator],
            schedulers=[scheduler_model, scheduler_discriminator],
            weighting=weighting,
            loss=loss,
        )

    def forward(self, x):
        """
        Forward pass.

        :param LabelTensor x: Input tensor.
        :return: The output of the neural network.
        :rtype: LabelTensor
        """
        return self.neural_net(x)

    def training_step(self, batch):
        """
        Solver training step, overridden to perform manual optimization.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The aggregated loss.
        :rtype: LabelTensor
        """
        # train model
        self.optimizer_model.instance.zero_grad()
        loss = super().training_step(batch)
        self.manual_backward(loss)
        self.optimizer_model.instance.step()
        self.scheduler_model.instance.step()

        # train discriminator
        self.optimizer_discriminator.instance.zero_grad()
        loss = super().training_step(batch)
        self.manual_backward(-loss)
        self.optimizer_discriminator.instance.step()
        self.scheduler_discriminator.instance.step()

        return loss

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        # Compute discriminator bets
        discriminator_bets = self.discriminator(samples)

        # Compute residual and multiply discriminator_bets
        residual = self.compute_residual(samples=samples, equation=equation)
        residual = residual * discriminator_bets

        # Compute competitive residual.
        loss_val = self._loss_fn(
            torch.zeros_like(residual, requires_grad=True),
            residual,
        )
        return loss_val

    def loss_data(self, input, target):
        """
        Compute the data loss for the PINN solver by evaluating the loss
        between the network's output and the true solution. This method should
        not be overridden, if not intentionally.

        :param input: The input to the neural network.
        :type input: LabelTensor
        :param target: The target to compare with the network's output.
        :type target: LabelTensor
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor
        """
        return self._loss_fn(self.forward(input), target)

    def configure_optimizers(self):
        """
        Optimizer configuration.

        :return: The optimizers and the schedulers
        :rtype: tuple[list[Optimizer], list[Scheduler]]
        """
        # If the problem is an InverseProblem, add the unknown parameters
        # to the parameters to be optimized
        self.optimizer_model.hook(self.neural_net.parameters())
        self.optimizer_discriminator.hook(self.discriminator.parameters())
        if isinstance(self.problem, InverseProblem):
            self.optimizer_model.instance.add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        self.scheduler_model.hook(self.optimizer_model)
        self.scheduler_discriminator.hook(self.optimizer_discriminator)
        return (
            [
                self.optimizer_model.instance,
                self.optimizer_discriminator.instance,
            ],
            [
                self.scheduler_model.instance,
                self.scheduler_discriminator.instance,
            ],
        )

    @property
    def neural_net(self):
        """
        The model.

        :return: The model.
        :rtype: torch.nn.Module
        """
        return self.models[0]

    @property
    def discriminator(self):
        """
        The discriminator.

        :return: The discriminator.
        :rtype: torch.nn.Module
        """
        return self.models[1]

    @property
    def optimizer_model(self):
        """
        The optimizer associated to the model.

        :return: The optimizer for the model.
        :rtype: Optimizer
        """
        return self.optimizers[0]

    @property
    def optimizer_discriminator(self):
        """
        The optimizer associated to the discriminator.

        :return: The optimizer for the discriminator.
        :rtype: Optimizer
        """
        return self.optimizers[1]

    @property
    def scheduler_model(self):
        """
        The scheduler associated to the model.

        :return: The scheduler for the model.
        :rtype: Scheduler
        """
        return self.schedulers[0]

    @property
    def scheduler_discriminator(self):
        """
        The scheduler associated to the discriminator.

        :return: The scheduler for the discriminator.
        :rtype: Scheduler
        """
        return self.schedulers[1]



"""Module for the ParametricProblem class."""

from abc import abstractmethod

from .abstract_problem import AbstractProblem


class ParametricProblem(AbstractProblem):
    """
    Class for defining parametric problems, where certain input variables are
    treated as parameters that can vary, allowing the model to adapt to
    different scenarios based on the chosen parameters.
    """

    @abstractmethod
    def parameter_domain(self):
        """
        The domain of the parameters of the problem.
        """

    @property
    def parameters(self):
        """
        Get the parameters of the problem.

        :return: The parameters of the problem.
        :rtype: list[str]
        """
        return self.parameter_domain.variables



"""Module for the Condition class."""

from .data_condition import DataCondition
from .domain_equation_condition import DomainEquationCondition
from .input_equation_condition import InputEquationCondition
from .input_target_condition import InputTargetCondition


class Condition:
    """
    The :class:`Condition` class is a core component of the PINA framework that
    provides a unified interface to define heterogeneous constraints that must
    be satisfied by a :class:`~pina.problem.abstract_problem.AbstractProblem`.

    It encapsulates all types of constraints - physical, boundary, initial, or
    data-driven - that the solver must satisfy during training. The specific
    behavior is inferred from the arguments passed to the constructor.

    Multiple types of conditions can be used within the same problem, allowing
    for a high degree of flexibility in defining complex problems.

    The :class:`Condition` class behavior specializes internally based on the
    arguments provided during instantiation. Depending on the specified keyword
    arguments, the class automatically selects the appropriate internal
    implementation.


    Available `Condition` types:

    - :class:`~pina.condition.input_target_condition.InputTargetCondition`:
      represents a supervised condition defined by both ``input`` and ``target``
      data. The model is trained to reproduce the ``target`` values given the
      ``input``. Supported data types include :class:`torch.Tensor`,
      :class:`~pina.label_tensor.LabelTensor`, :class:`~pina.graph.Graph`, or
      :class:`~torch_geometric.data.Data`.
      The class automatically selects the appropriate implementation based on
      the types of ``input`` and ``target``.

    - :class:`~pina.condition.domain_equation_condition.DomainEquationCondition`
      : represents a general physics-informed condition defined by a ``domain``
      and an ``equation``. The model learns to minimize the equation residual
      through evaluations performed at points sampled from the specified domain.

    - :class:`~pina.condition.input_equation_condition.InputEquationCondition`:
      represents a general physics-informed condition defined by ``input``
      points and an ``equation``. The model learns to minimize the equation
      residual through evaluations performed at the provided ``input``.
      Supported data types for the ``input`` include
      :class:`~pina.label_tensor.LabelTensor` or :class:`~pina.graph.Graph`.
      The class automatically selects the appropriate implementation based on
      the types of the ``input``.

    - :class:`~pina.condition.data_condition.DataCondition`: represents an
      unsupervised, data-driven condition defined by the ``input`` only.
      The model is trained using a custom unsupervised loss determined by the
      chosen :class:`~pina.solver.solver.SolverInterface`, while leveraging the
      provided data during training. Optional ``conditional_variables`` can be
      specified when the model depends on additional parameters.
      Supported data types include :class:`torch.Tensor`,
      :class:`~pina.label_tensor.LabelTensor`, :class:`~pina.graph.Graph`, or
      :class:`~torch_geometric.data.Data`.
      The class automatically selects the appropriate implementation based on
      the type of the ``input``.

    .. note::

        The user should always instantiate :class:`Condition` directly, without
        manually creating subclass instances. Please refer to the specific
        :class:`Condition` classes for implementation details.

    :Example:

    >>> from pina import Condition

    >>> # Example of InputTargetCondition signature
    >>> condition = Condition(input=input, target=target)

    >>> # Example of DomainEquationCondition signature
    >>> condition = Condition(domain=domain, equation=equation)

    >>> # Example of InputEquationCondition signature
    >>> condition = Condition(input=input, equation=equation)

    >>> # Example of DataCondition signature
    >>> condition = Condition(input=data, conditional_variables=cond_vars)
    """

    # Combine all possible keyword arguments from the different Condition types
    __slots__ = list(
        set(
            InputTargetCondition.__slots__
            + InputEquationCondition.__slots__
            + DomainEquationCondition.__slots__
            + DataCondition.__slots__
        )
    )

    def __new__(cls, *args, **kwargs):
        """
        Instantiate the appropriate :class:`Condition` object based on the
        keyword arguments passed.

        :param tuple args: The positional arguments (should be empty).
        :param dict kwargs: The keyword arguments corresponding to the
            parameters of the specific :class:`Condition` type to instantiate.
        :raises ValueError: If unexpected positional arguments are provided.
        :raises ValueError: If the keyword arguments are invalid.
        :return: The appropriate :class:`Condition` object.
        :rtype: ConditionInterface
        """
        # Check keyword arguments
        if len(args) != 0:
            raise ValueError(
                "Condition takes only the following keyword "
                f"arguments: {Condition.__slots__}."
            )

        # Class specialization based on keyword arguments
        sorted_keys = sorted(kwargs.keys())

        # Input - Target Condition
        if sorted_keys == sorted(InputTargetCondition.__slots__):
            return InputTargetCondition(**kwargs)

        # Input - Equation Condition
        if sorted_keys == sorted(InputEquationCondition.__slots__):
            return InputEquationCondition(**kwargs)

        # Domain - Equation Condition
        if sorted_keys == sorted(DomainEquationCondition.__slots__):
            return DomainEquationCondition(**kwargs)

        # Data Condition
        if (
            sorted_keys == sorted(DataCondition.__slots__)
            or sorted_keys[0] == DataCondition.__slots__[0]
        ):
            return DataCondition(**kwargs)

        # Invalid keyword arguments
        raise ValueError(f"Invalid keyword arguments {kwargs.keys()}.")



"""Module for the Ellipsoid Domain."""

import torch
from .domain_interface import DomainInterface
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class EllipsoidDomain(DomainInterface):
    """
    Implementation of the ellipsoid domain.
    """

    def __init__(self, ellipsoid_dict, sample_surface=False):
        """
        Initialization of the :class:`EllipsoidDomain` class.

        :param dict ellipsoid_dict: A dictionary where the keys are the variable
            names and the values are the domain extrema.
        :param bool sample_surface: A flag to choose the sampling strategy.
            If ``True``, samples are taken from the surface of the ellipsoid.
            If ``False``, samples are taken from the interior of the ellipsoid.
            Default is ``False``.
        :raises TypeError: If the input dictionary is not correctly formatted.

        .. warning::
            Sampling for dimensions greater or equal to 10 could result in a
            shrinkage of the ellipsoid, which degrades the quality of the
            samples. For dimensions higher than 10, see the following reference.

        .. seealso::
            **Original reference**: Dezert, Jean, and Musso, Christian.
            *An efficient method for generating points uniformly distributed
            in hyperellipsoids.*
            Proceedings of the Workshop on Estimation, Tracking and Fusion:
            A Tribute to Yaakov Bar-Shalom. 2001.

        :Example:
            >>> spatial_domain = Ellipsoid({'x':[-1, 1], 'y':[-1,1]})
        """
        self.fixed_ = {}
        self.range_ = {}
        self._centers = None
        self._axis = None

        # checking consistency
        check_consistency(sample_surface, bool)
        self._sample_surface = sample_surface

        for k, v in ellipsoid_dict.items():
            if isinstance(v, (int, float)):
                self.fixed_[k] = v
            elif isinstance(v, (list, tuple)) and len(v) == 2:
                self.range_[k] = v
            else:
                raise TypeError

        # perform operation only for not fixed variables (if any)

        if self.range_:
            # convert dict vals to torch [dim, 2] matrix
            list_dict_vals = list(self.range_.values())
            tmp = torch.tensor(list_dict_vals, dtype=torch.float)

            # get the ellipsoid center
            normal_basis = torch.eye(len(list_dict_vals))
            centers = torch.diag(normal_basis * tmp.mean(axis=1))

            # get the ellipsoid axis
            ellipsoid_axis = (tmp - centers.reshape(-1, 1))[:, -1]

            # save elipsoid axis and centers as dict
            self._centers = dict(zip(self.range_.keys(), centers.tolist()))
            self._axis = dict(zip(self.range_.keys(), ellipsoid_axis.tolist()))

    @property
    def sample_modes(self):
        """
        List of available sampling modes.

        :return: List of available sampling modes.
        :rtype: list[str]
        """
        return ["random"]

    @property
    def variables(self):
        """
        List of variables of the domain.

        :return: List of variables of the domain.
        :rtype: list[str]
        """
        return sorted(list(self.fixed_.keys()) + list(self.range_.keys()))

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the ellipsoid.

        :param LabelTensor point: Point to be checked.
        :param bool check_border: If ``True``, the border is considered inside
            the ellipsoid. Default is ``False``.
        :raises ValueError: If the labels of the point are different from those
            passed in the ``__init__`` method.
        :return: ``True`` if the point is inside the domain,
            ``False`` otherwise.
        :rtype: bool

        .. note::
            When ``sample_surface=True`` in the ``__init__`` method, this method
            checks only those points lying on the surface of the ellipsoid.
        """
        # small check that point is labeltensor
        check_consistency(point, LabelTensor)

        # get axis ellipse as tensors
        list_dict_vals = list(self._axis.values())
        tmp = torch.tensor(list_dict_vals, dtype=torch.float)
        ax_sq = LabelTensor(tmp.reshape(1, -1) ** 2, self.variables)

        # get centers ellipse as tensors
        list_dict_vals = list(self._centers.values())
        tmp = torch.tensor(list_dict_vals, dtype=torch.float)
        centers = LabelTensor(tmp.reshape(1, -1), self.variables)

        if not all(i in ax_sq.labels for i in point.labels):
            raise ValueError(
                "point labels different from constructor"
                f" dictionary labels. Got {point.labels},"
                f" expected {ax_sq.labels}."
            )

        # point square + shift center
        point_sq = (point - centers).pow(2)
        point_sq.labels = point.labels

        # calculate ellispoid equation
        eqn = torch.sum(point_sq.extract(ax_sq.labels) / ax_sq) - 1.0

        # if we have sampled only the surface, we check that the
        # point is inside the surface border only
        if self._sample_surface:
            return torch.allclose(eqn, torch.zeros_like(eqn))

        # otherwise we check the ellipse
        if check_border:
            return bool(eqn <= 0)

        return bool(eqn < 0)

    def _sample_range(self, n, mode, variables):
        """
        Rescale the samples to fit within the specified bounds.

        :param int n: Number of points to sample.
        :param str mode: Sampling method. Default is ``random``.
        :param list[str] variables: variables whose samples must be rescaled.
        :return: Rescaled sample points.
        :rtype: torch.Tensor
        """

        # =============== For Developers ================ #
        #
        # The sampling startegy used is fairly simple.
        # For all `mode`s first we sample from the unit
        # sphere and then we scale and shift according
        # to self._axis.values() and self._centers.values().
        #
        # =============================================== #

        # get dimension
        dim = len(variables)

        # get values center
        pairs_center = [
            (k, v) for k, v in self._centers.items() if k in variables
        ]
        _, values_center = map(list, zip(*pairs_center))
        values_center = torch.tensor(values_center)

        # get values axis
        pairs_axis = [(k, v) for k, v in self._axis.items() if k in variables]
        _, values_axis = map(list, zip(*pairs_axis))
        values_axis = torch.tensor(values_axis)

        # Sample in the unit sphere
        if mode == "random":
            # 1. Sample n points from the surface of a unit sphere
            # 2. Scale each dimension using torch.rand()
            #    (a random number between 0-1) so that it lies within
            #    the sphere, only if self._sample_surface=False
            # 3. Multiply with self._axis.values() to make it ellipsoid
            # 4. Shift the mean of the ellipse by adding self._centers.values()

            # step 1.
            pts = torch.randn(size=(n, dim))
            pts = pts / torch.linalg.norm(pts, axis=-1).view((n, 1))
            if not self._sample_surface:  # step 2.
                scale = torch.rand((n, 1))
                pts = pts * scale

            # step 3. and 4.
            pts *= values_axis
            pts += values_center

        return pts

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: Number of points to sample.
        :param str mode: Sampling method. Default is ``random``.
            Available modes: random sampling, ``random``.
        :param list[str] variables: variables to be sampled. Default is ``all``.
        :raises NotImplementedError: If the sampling mode is not implemented.
        :return: Sampled points.
        :rtype: LabelTensor

        :Example:
            >>> ellipsoid = Ellipsoid({'x':[1, 0], 'y':1})
            >>> ellipsoid.sample(n=6)
                tensor([[0.4872, 1.0000],
                        [0.2977, 1.0000],
                        [0.0422, 1.0000],
                        [0.6431, 1.0000],
                        [0.7272, 1.0000],
                        [0.8326, 1.0000]])
        """

        def _Nd_sampler(n, mode, variables):
            """
            Sample all variables together.

            :param int n: Number of points to sample.
            :param str mode: Sampling method.
            :param list[str] variables: variables to be sampled.
            :return: Sampled points.
            :rtype: list[LabelTensor]
            """
            pairs = [(k, v) for k, v in self.range_.items() if k in variables]
            keys, _ = map(list, zip(*pairs))

            result = self._sample_range(n, mode, keys)
            result = result.as_subclass(LabelTensor)
            result.labels = keys

            for variable in variables:
                if variable in self.fixed_:
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(
                        result.shape[0], 1
                    )
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    result = result.append(pts_variable, mode="std")
            return result

        def _single_points_sample(n, variables):
            """
            Sample a single point in one dimension.

            :param int n: Number of points to sample.
            :param list[str] variables: variables to be sampled.
            :return: Sampled points.
            :rtype: list[torch.Tensor]
            """
            tmp = []
            for variable in variables:
                if variable in self.fixed_:
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(n, 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]
                    tmp.append(pts_variable)

            result = tmp[0]
            for i in tmp[1:]:
                result = result.append(i, mode="std")

            return result

        if variables == "all":
            variables = self.variables
        elif isinstance(variables, (list, tuple)):
            variables = sorted(variables)

        if self.fixed_ and (not self.range_):
            return _single_points_sample(n, variables).extract(variables)

        if mode in self.sample_modes:
            return _Nd_sampler(n, mode, variables).extract(variables)

        raise NotImplementedError(f"mode={mode} is not implemented.")



"""Module for Self-Adaptive Weighting class."""

import torch
from .weighting_interface import WeightingInterface


class SelfAdaptiveWeighting(WeightingInterface):
    """
    A self-adaptive weighting scheme to tackle the imbalance among the loss
    components. This formulation equalizes the gradient norms of the losses,
    preventing bias toward any particular term during training.

    .. seealso::

        **Original reference**:
        Wang, S., Sankaran, S., Stinis., P., Perdikaris, P. (2025).
        *Simulating Three-dimensional Turbulence with Physics-informed Neural
        Networks*.
        DOI: `arXiv preprint arXiv:2507.08972.
        <https://arxiv.org/abs/2507.08972>`_

    """

    def __init__(self, update_every_n_epochs=1):
        """
        Initialization of the :class:`SelfAdaptiveWeighting` class.

        :param int update_every_n_epochs: The number of training epochs between
            weight updates. If set to 1, the weights are updated at every epoch.
            Default is 1.
        """
        super().__init__(update_every_n_epochs=update_every_n_epochs)

    def weights_update(self, losses):
        """
        Update the weighting scheme based on the given losses.

        :param dict losses: The dictionary of losses.
        :return: The updated weights.
        :rtype: dict
        """
        # Define a dictionary to store the norms of the gradients
        losses_norm = {}

        # Compute the gradient norms for each loss component
        for condition, loss in losses.items():
            loss.backward(retain_graph=True)
            grads = torch.cat(
                [p.grad.flatten() for p in self.solver.model.parameters()]
            )
            losses_norm[condition] = grads.norm()

        # Update the weights
        return {
            condition: sum(losses_norm.values()) / losses_norm[condition]
            for condition in losses
        }



"""Module for Neural Tangent Kernel Class"""

import torch
from .weighting_interface import WeightingInterface
from ..utils import check_consistency, in_range


class NeuralTangentKernelWeighting(WeightingInterface):
    """
    A neural tangent kernel scheme for weighting different losses to
    boost the convergence.

    .. seealso::

        **Original reference**: Wang, Sifan, Xinling Yu, and
        Paris Perdikaris. *When and why PINNs fail to train:
        A neural tangent kernel perspective*. Journal of
        Computational Physics 449 (2022): 110768.
        DOI: `10.1016 <https://doi.org/10.1016/j.jcp.2021.110768>`_.

    """

    def __init__(self, update_every_n_epochs=1, alpha=0.5):
        """
        Initialization of the :class:`NeuralTangentKernelWeighting` class.

        :param int update_every_n_epochs: The number of training epochs between
            weight updates. If set to 1, the weights are updated at every epoch.
            Default is 1.
        :param float alpha: The alpha parameter.
        :raises ValueError: If ``alpha`` is not between 0 and 1 (inclusive).
        """
        super().__init__(update_every_n_epochs=update_every_n_epochs)

        # Check consistency
        check_consistency(alpha, float)
        if not in_range(alpha, [0, 1], strict=False):
            raise ValueError("alpha must be in range (0, 1).")

        # Initialize parameters
        self.alpha = alpha
        self.weights = {}

    def weights_update(self, losses):
        """
        Update the weighting scheme based on the given losses.

        :param dict losses: The dictionary of losses.
        :return: The updated weights.
        :rtype: dict
        """
        # Define a dictionary to store the norms of the gradients
        losses_norm = {}

        # Compute the gradient norms for each loss component
        for condition, loss in losses.items():
            loss.backward(retain_graph=True)
            grads = torch.cat(
                [p.grad.flatten() for p in self.solver.model.parameters()]
            )
            losses_norm[condition] = grads.norm()

        # Update the weights
        return {
            condition: self.alpha * self.last_saved_weights().get(condition, 1)
            + (1 - self.alpha)
            * losses_norm[condition]
            / sum(losses_norm.values())
            for condition in losses
        }



"""Module for residual blocks and enhanced linear layers."""

import torch
from torch import nn
from ...utils import check_consistency


class ResidualBlock(nn.Module):
    """
    Residual block class.

    .. seealso::

        **Original reference**: He, Kaiming, et al.
        *Deep residual learning for image recognition.*
        Proceedings of the IEEE conference on computer vision and pattern
        recognition. 2016.
        DOI: `<https://arxiv.org/pdf/1512.03385.pdf>`_.
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        hidden_dim,
        spectral_norm=False,
        activation=torch.nn.ReLU(),
    ):
        """
        Initialization of the :class:`ResidualBlock` class.

        :param int input_dim: The input dimension.
        :param int output_dim: The output dimension.
        :param int hidden_dim: The hidden dimension.
        :param bool spectral_norm: If ``True``, the spectral normalization is
            applied to the feedforward layers. Default is ``False``.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.ReLU`.

        """
        super().__init__()
        # check consistency
        check_consistency(spectral_norm, bool)
        check_consistency(input_dim, int)
        check_consistency(output_dim, int)
        check_consistency(hidden_dim, int)
        check_consistency(activation, torch.nn.Module)

        # assign variables
        self._spectral_norm = spectral_norm
        self._input_dim = input_dim
        self._output_dim = output_dim
        self._hidden_dim = hidden_dim
        self._activation = activation

        # create layers
        self._l1 = self._spect_norm(nn.Linear(input_dim, hidden_dim))
        self._l2 = self._spect_norm(nn.Linear(hidden_dim, output_dim))
        self._l3 = self._spect_norm(nn.Linear(input_dim, output_dim))

    def forward(self, x):
        """
        Forward pass.

        :param torch.Tensor x: The input tensor.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        y = self._activation(self._l1(x))
        y = self._l2(y)
        x = self._l3(x)
        return y + x

    def _spect_norm(self, x):
        """
        Perform spectral normalization on the network layers.

        :param torch.nn.Module x: A :class:`torch.nn.Linear` layer.
        :return: The spectral norm of the layer
        :rtype: torch.nn.Module
        """
        return nn.utils.spectral_norm(x) if self._spectral_norm else x


class EnhancedLinear(torch.nn.Module):
    """
    Enhanced Linear layer class.

    This class is a wrapper for enhancing a linear layer with activation and/or
    dropout.
    """

    def __init__(self, layer, activation=None, dropout=None):
        """
        Initialization of the :class:`EnhancedLinear` class.

        :param torch.nn.Module layer: The linear layer to be enhanced.
        :param torch.nn.Module activation: The activation function. Default is
            ``None``.
        :param float dropout: The dropout probability. Default is ``None``.

        :Example:

            >>> linear_layer = torch.nn.Linear(10, 20)
            >>> activation = torch.nn.ReLU()
            >>> dropout_prob = 0.5
            >>> enhanced_linear = EnhancedLinear(
            ...     linear_layer,
            ...     activation,
            ...     dropout_prob
            ... )
        """
        super().__init__()

        # check consistency
        check_consistency(layer, nn.Module)
        if activation is not None:
            check_consistency(activation, nn.Module)
        if dropout is not None:
            check_consistency(dropout, float)

        # assign forward
        if (dropout is None) and (activation is None):
            self._model = torch.nn.Sequential(layer)

        elif (dropout is None) and (activation is not None):
            self._model = torch.nn.Sequential(layer, activation)

        elif (dropout is not None) and (activation is None):
            self._model = torch.nn.Sequential(layer, self._drop(dropout))

        elif (dropout is not None) and (activation is not None):
            self._model = torch.nn.Sequential(
                layer, activation, self._drop(dropout)
            )

    def forward(self, x):
        """
        Forward pass.

        :param torch.Tensor x: The input tensor.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        return self._model(x)

    def _drop(self, p):
        """
        Apply dropout with probability p.

        :param float p: Dropout probability.
        :return: Dropout layer with the specified probability.
        :rtype: torch.nn.Dropout
        """
        return torch.nn.Dropout(p)



"""Module to define equations and systems of equations."""

__all__ = [
    "SystemEquation",
    "Equation",
    "FixedValue",
    "FixedGradient",
    "FixedFlux",
    "FixedLaplacian",
    "Laplace",
    "Advection",
    "AllenCahn",
    "DiffusionReaction",
    "Helmholtz",
    "Poisson",
]

from .equation import Equation
from .equation_factory import (
    FixedFlux,
    FixedGradient,
    FixedLaplacian,
    FixedValue,
    Laplace,
    Advection,
    AllenCahn,
    DiffusionReaction,
    Helmholtz,
    Poisson,
)
from .system_equation import SystemEquation



"""
This module contains condition classes for supervised learning tasks.
"""

import torch
from torch_geometric.data import Data
from ..label_tensor import LabelTensor
from ..graph import Graph
from .condition_interface import ConditionInterface


class InputTargetCondition(ConditionInterface):
    """
    The :class:`InputTargetCondition` class represents a supervised condition
    defined by both ``input`` and ``target`` data. The model is trained to
    reproduce the ``target`` values given the ``input``. Supported data types
    include :class:`torch.Tensor`, :class:`~pina.label_tensor.LabelTensor`,
    :class:`~pina.graph.Graph`, or :class:`~torch_geometric.data.Data`.

    The class automatically selects the appropriate implementation based on
    the types of ``input`` and ``target``. Depending on whether the ``input``
    and ``target`` are tensors or graph-based data, one of the following
    specialized subclasses is instantiated:

    - :class:`TensorInputTensorTargetCondition`: For cases where both ``input``
      and ``target`` data are either :class:`torch.Tensor` or
      :class:`~pina.label_tensor.LabelTensor`.

    - :class:`TensorInputGraphTargetCondition`: For cases where ``input`` is
      either a :class:`torch.Tensor` or :class:`~pina.label_tensor.LabelTensor`
      and ``target`` is either a :class:`~pina.graph.Graph` or a
      :class:`torch_geometric.data.Data`.

    - :class:`GraphInputTensorTargetCondition`: For cases where ``input`` is
      either a :class:`~pina.graph.Graph` or :class:`torch_geometric.data.Data`
      and ``target`` is either a :class:`torch.Tensor` or a
      :class:`~pina.label_tensor.LabelTensor`.

    - :class:`GraphInputGraphTargetCondition`: For cases where both ``input``
      and ``target`` are either :class:`~pina.graph.Graph` or
      :class:`torch_geometric.data.Data`.

    :Example:

    >>> from pina import Condition, LabelTensor
    >>> from pina.graph import Graph
    >>> import torch

    >>> pos = LabelTensor(torch.randn(100, 2), labels=["x", "y"])
    >>> edge_index = torch.randint(0, 100, (2, 300))
    >>> graph = Graph(pos=pos, edge_index=edge_index)

    >>> input = LabelTensor(torch.randn(100, 2), labels=["x", "y"])
    >>> condition = Condition(input=input, target=graph)
    """

    # Available input and target data types
    __slots__ = ["input", "target"]
    _avail_input_cls = (torch.Tensor, LabelTensor, Data, Graph, list, tuple)
    _avail_output_cls = (torch.Tensor, LabelTensor, Data, Graph, list, tuple)

    def __new__(cls, input, target):
        """
        Instantiate the appropriate subclass of :class:`InputTargetCondition`
        based on the types of both ``input`` and ``target`` data.

        :param input: The input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param target: The target data for the condition.
        :type target: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :return: The subclass of InputTargetCondition.
        :rtype: pina.condition.input_target_condition.
            TensorInputTensorTargetCondition |
            pina.condition.input_target_condition.
            TensorInputGraphTargetCondition |
            pina.condition.input_target_condition.
            GraphInputTensorTargetCondition |
            pina.condition.input_target_condition.GraphInputGraphTargetCondition

        :raises ValueError: If ``input`` and/or ``target`` are not of type
            :class:`torch.Tensor`, :class:`~pina.label_tensor.LabelTensor`,
            :class:`~pina.graph.Graph`, or :class:`~torch_geometric.data.Data`.
        """
        if cls != InputTargetCondition:
            return super().__new__(cls)

        # Tensor - Tensor
        if isinstance(input, (torch.Tensor, LabelTensor)) and isinstance(
            target, (torch.Tensor, LabelTensor)
        ):
            subclass = TensorInputTensorTargetCondition
            return subclass.__new__(subclass, input, target)

        # Tensor - Graph
        if isinstance(input, (torch.Tensor, LabelTensor)) and isinstance(
            target, (Graph, Data, list, tuple)
        ):
            cls._check_graph_list_consistency(target)
            subclass = TensorInputGraphTargetCondition
            return subclass.__new__(subclass, input, target)

        # Graph - Tensor
        if isinstance(input, (Graph, Data, list, tuple)) and isinstance(
            target, (torch.Tensor, LabelTensor)
        ):
            cls._check_graph_list_consistency(input)
            subclass = GraphInputTensorTargetCondition
            return subclass.__new__(subclass, input, target)

        # Graph - Graph
        if isinstance(input, (Graph, Data, list, tuple)) and isinstance(
            target, (Graph, Data, list, tuple)
        ):
            cls._check_graph_list_consistency(input)
            cls._check_graph_list_consistency(target)
            subclass = GraphInputGraphTargetCondition
            return subclass.__new__(subclass, input, target)

        # If the input and/or target are not of the correct type raise an error
        raise ValueError(
            "Invalid input | target types."
            "Please provide either torch_geometric.data.Data, Graph, "
            "LabelTensor or torch.Tensor objects."
        )

    def __init__(self, input, target):
        """
        Initialization of the :class:`InputTargetCondition` class.

        :param input: The input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param target: The target data for the condition.
        :type target: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]

        .. note::

            If either ``input`` or ``target`` is a list of
            :class:`~pina.graph.Graph` or :class:`~torch_geometric.data.Data`
            objects, all elements in the list must share the same structure,
            with matching keys and consistent data types.
        """
        super().__init__()
        self._check_input_target_len(input, target)
        self.input = input
        self.target = target

    @staticmethod
    def _check_input_target_len(input, target):
        """
        Check that the length of the input and target lists are the same.

        :param input: The input data.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param target: The target data.
        :type target: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :raises ValueError: If the lengths of the input and target lists do not
            match.
        """
        if isinstance(input, (Graph, Data)) or isinstance(
            target, (Graph, Data)
        ):
            return

        # Raise an error if the lengths of the input and target do not match
        if len(input) != len(target):
            raise ValueError(
                "The input and target lists must have the same length."
            )


class TensorInputTensorTargetCondition(InputTargetCondition):
    """
    Specialization of the :class:`InputTargetCondition` class for the case where
    both ``input`` and ``target`` are :class:`torch.Tensor` or
    :class:`~pina.label_tensor.LabelTensor` objects.
    """


class TensorInputGraphTargetCondition(InputTargetCondition):
    """
    Specialization of the :class:`InputTargetCondition` class for the case where
    ``input`` is either a :class:`torch.Tensor` or a
    :class:`~pina.label_tensor.LabelTensor` object and ``target`` is either a
    :class:`~pina.graph.Graph` or a :class:`torch_geometric.data.Data` object.
    """


class GraphInputTensorTargetCondition(InputTargetCondition):
    """
    Specialization of the :class:`InputTargetCondition` class for the case where
    ``input`` is either a :class:`~pina.graph.Graph` or
    :class:`torch_geometric.data.Data` object and ``target`` is either a
    :class:`torch.Tensor` or a :class:`~pina.label_tensor.LabelTensor` object.
    """


class GraphInputGraphTargetCondition(InputTargetCondition):
    """
    Specialization of the :class:`InputTargetCondition` class for the case where
    both ``input`` and ``target`` are either :class:`~pina.graph.Graph` or
    :class:`torch_geometric.data.Data` objects.
    """



"""Module for the Physics-Informed Neural Network Interface."""

from abc import ABCMeta, abstractmethod
import warnings
import torch

from ...utils import custom_warning_format
from ..supervised_solver import SupervisedSolverInterface
from ...condition import (
    InputTargetCondition,
    InputEquationCondition,
    DomainEquationCondition,
)

# set the warning for torch >= 2.8 compile
warnings.formatwarning = custom_warning_format
warnings.filterwarnings("always", category=UserWarning)


class PINNInterface(SupervisedSolverInterface, metaclass=ABCMeta):
    """
    Base class for Physics-Informed Neural Network (PINN) solvers, implementing
    the :class:`~pina.solver.solver.SolverInterface` class.

    The `PINNInterface` class can be used to define PINNs that work with one or
    multiple optimizers and/or models. By default, it is compatible with
    problems defined by :class:`~pina.problem.abstract_problem.AbstractProblem`,
    and users can choose the problem type the solver is meant to address.
    """

    accepted_conditions_types = (
        InputTargetCondition,
        InputEquationCondition,
        DomainEquationCondition,
    )

    def __init__(self, **kwargs):
        """
        Initialization of the :class:`PINNInterface` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param kwargs: Additional keyword arguments to be passed to the
            :class:`~pina.solver.supervised_solver.SupervisedSolverInterface`
            class.
        """
        kwargs["use_lt"] = True
        super().__init__(**kwargs)

        # current condition name
        self.__metric = None

    def setup(self, stage):
        """
        Setup method executed at the beginning of training and testing.

        This method compiles the model only if the installed torch version
        is earlier than 2.8, due to known issues with later versions
        (see https://github.com/mathLab/PINA/issues/621).

        .. warning::
            For torch >= 2.8, compilation is disabled. Forcing compilation
            on these versions may cause runtime errors or unstable behavior.

        :param str stage: The current stage of the training process
            (e.g., ``fit``, ``validate``, ``test``, ``predict``).
        :return: The result of the parent class ``setup`` method.
        :rtype: Any
        """
        # Override the compilation, compiling only for torch < 2.8, see
        # related issue at https://github.com/mathLab/PINA/issues/621
        if torch.__version__ >= "2.8":
            self.trainer.compile = False
            warnings.warn(
                "Compilation is disabled for torch >= 2.8. "
                "Forcing compilation may cause runtime errors or instability.",
                UserWarning,
            )
        return super().setup(stage)

    def optimization_cycle(self, batch, loss_residuals=None):
        """
        The optimization cycle for the PINN solver.

        This method allows to call `_run_optimization_cycle` with the physics
        loss as argument, thus distinguishing the training step from the
        validation and test steps.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        # which losses to use
        if loss_residuals is None:
            loss_residuals = self.loss_phys
        # compute optimization cycle
        condition_loss = {}
        for condition_name, points in batch:
            self.__metric = condition_name
            # if equations are passed
            if "target" not in points:
                input_pts = points["input"]
                condition = self.problem.conditions[condition_name]
                loss = loss_residuals(
                    input_pts.requires_grad_(), condition.equation
                )
            # if data are passed
            else:
                input_pts = points["input"]
                output_pts = points["target"]
                loss = self.loss_data(
                    input=input_pts.requires_grad_(), target=output_pts
                )
            # append loss
            condition_loss[condition_name] = loss
        return condition_loss

    @torch.set_grad_enabled(True)
    def validation_step(self, batch):
        """
        The validation step for the PINN solver. It returns the average residual
        computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The loss of the validation step.
        :rtype: torch.Tensor
        """
        return super().validation_step(
            batch, loss_residuals=self._residual_loss
        )

    @torch.set_grad_enabled(True)
    def test_step(self, batch):
        """
        The test step for the PINN solver. It returns the average residual
        computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The loss of the test step.
        :rtype: torch.Tensor
        """
        return super().test_step(batch, loss_residuals=self._residual_loss)

    def loss_data(self, input, target):
        """
        Compute the data loss for the PINN solver by evaluating the loss
        between the network's output and the true solution. This method should
        be overridden by the derived class.

        :param LabelTensor input: The input to the neural network.
        :param LabelTensor target: The target to compare with the
            network's output.
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor
        :raises NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError(
            "PINN is being used in a supervised learning context, but the "
            "'loss_data' method has not been implemented. "
        )

    @abstractmethod
    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation. This method must be overridden in
        subclasses. It distinguishes different types of PINN solvers.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """

    def compute_residual(self, samples, equation):
        """
        Compute the residuals of the equation.

        :param LabelTensor samples: The samples to evaluate the loss.
        :param EquationInterface equation: The governing equation.
        :return: The residual of the solution of the model.
        :rtype: LabelTensor
        """
        residual = equation.residual(
            samples, self.forward(samples), self._params
        )
        return residual

    def _residual_loss(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation. This method should never be overridden
        by the user, if not intentionally,
        since it is used internally to compute validation loss.


        :param LabelTensor samples: The samples to evaluate the loss.
        :param EquationInterface equation: The governing equation.
        :return: The residual loss.
        :rtype: torch.Tensor
        """
        residuals = self.compute_residual(samples, equation)
        return self._loss_fn(residuals, torch.zeros_like(residuals))

    @property
    def current_condition_name(self):
        """
        The current condition name.

        :return: The current condition name.
        :rtype: str
        """
        return self.__metric



"""Module for the Trainer."""

import sys
import warnings
import torch
import lightning
from .utils import check_consistency, custom_warning_format
from .data import PinaDataModule
from .solver import SolverInterface, PINNInterface

# set the warning for compile options
warnings.formatwarning = custom_warning_format
warnings.filterwarnings("always", category=UserWarning)


class Trainer(lightning.pytorch.Trainer):
    """
    PINA custom Trainer class to extend the standard Lightning functionality.

    This class enables specific features or behaviors required by the PINA
    framework. It modifies the standard
    :class:`lightning.pytorch.Trainer <lightning.pytorch.trainer.trainer.Trainer>`
    class to better support the training process in PINA.
    """

    def __init__(
        self,
        solver,
        batch_size=None,
        train_size=1.0,
        test_size=0.0,
        val_size=0.0,
        compile=None,
        repeat=None,
        automatic_batching=None,
        num_workers=None,
        pin_memory=None,
        shuffle=None,
        **kwargs,
    ):
        """
        Initialization of the :class:`Trainer` class.

        :param SolverInterface solver: A
            :class:`~pina.solver.solver.SolverInterface` solver used to solve a
            :class:`~pina.problem.abstract_problem.AbstractProblem`.
        :param int batch_size: The number of samples per batch to load.
            If ``None``, all samples are loaded and data is not batched.
            Default is ``None``.
        :param float train_size: The percentage of elements to include in the
            training dataset. Default is ``1.0``.
        :param float test_size: The percentage of elements to include in the
            test dataset. Default is ``0.0``.
        :param float val_size: The percentage of elements to include in the
            validation dataset. Default is ``0.0``.
        :param bool compile: If ``True``, the model is compiled before training.
            Default is ``False``. For Windows users, it is always disabled. Not
            supported for python version greater or equal than 3.14.
        :param bool repeat: Whether to repeat the dataset data in each
            condition during training. For further details, see the
            :class:`~pina.data.data_module.PinaDataModule` class. Default is
            ``False``.
        :param bool automatic_batching: If ``True``, automatic PyTorch batching
            is performed, otherwise the items are retrieved from the dataset
            all at once. For further details, see the
            :class:`~pina.data.data_module.PinaDataModule` class. Default is
            ``False``.
        :param int num_workers: The number of worker threads for data loading.
            Default is ``0`` (serial loading).
        :param bool pin_memory: Whether to use pinned memory for faster data
            transfer to GPU. Default is ``False``.
        :param bool shuffle: Whether to shuffle the data during training.
            Default is ``True``.
        :param dict kwargs: Additional keyword arguments that specify the
            training setup. These can be selected from the `pytorch-lightning
            Trainer API
            <https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api>`_.
        """
        # check consistency for init types
        self._check_input_consistency(
            solver=solver,
            train_size=train_size,
            test_size=test_size,
            val_size=val_size,
            repeat=repeat,
            automatic_batching=automatic_batching,
            compile=compile,
        )
        pin_memory, num_workers, shuffle, batch_size = (
            self._check_consistency_and_set_defaults(
                pin_memory, num_workers, shuffle, batch_size
            )
        )

        # inference mode set to false when validating/testing PINNs otherwise
        # gradient is not tracked and optimization_cycle fails
        if isinstance(solver, PINNInterface):
            kwargs["inference_mode"] = False

        # Logging depends on the batch size, when batch_size is None then
        # log_every_n_steps should be zero
        if batch_size is None:
            kwargs["log_every_n_steps"] = 0
        else:
            kwargs.setdefault("log_every_n_steps", 50)  # default for lightning

        # Setting default kwargs, overriding lightning defaults
        kwargs.setdefault("enable_progress_bar", True)

        super().__init__(**kwargs)

        # checking compilation and automatic batching
        # compilation disabled for Windows and for Python 3.14+
        if (
            compile is None
            or sys.platform == "win32"
            or sys.version_info >= (3, 14)
        ):
            compile = False
            warnings.warn(
                "Compilation is disabled for Python 3.14+ and for Windows.",
                UserWarning,
            )

        repeat = repeat if repeat is not None else False

        automatic_batching = (
            automatic_batching if automatic_batching is not None else False
        )

        # set attributes
        self.compile = compile
        self.solver = solver
        self.batch_size = batch_size
        self._move_to_device()
        self.data_module = None
        self._create_datamodule(
            train_size=train_size,
            test_size=test_size,
            val_size=val_size,
            batch_size=batch_size,
            repeat=repeat,
            automatic_batching=automatic_batching,
            pin_memory=pin_memory,
            num_workers=num_workers,
            shuffle=shuffle,
        )

        # logging
        self.logging_kwargs = {
            "sync_dist": bool(
                len(self._accelerator_connector._parallel_devices) > 1
            ),
            "on_step": bool(kwargs["log_every_n_steps"] > 0),
            "prog_bar": bool(kwargs["enable_progress_bar"]),
            "on_epoch": True,
        }

    def _move_to_device(self):
        """
        Moves the ``unknown_parameters`` of an instance of
        :class:`~pina.problem.abstract_problem.AbstractProblem` to the
        :class:`Trainer` device.
        """
        device = self._accelerator_connector._parallel_devices[0]
        # move parameters to device
        pb = self.solver.problem
        if hasattr(pb, "unknown_parameters"):
            for key in pb.unknown_parameters:
                pb.unknown_parameters[key] = torch.nn.Parameter(
                    pb.unknown_parameters[key].data.to(device)
                )

    def _create_datamodule(
        self,
        train_size,
        test_size,
        val_size,
        batch_size,
        repeat,
        automatic_batching,
        pin_memory,
        num_workers,
        shuffle,
    ):
        """
        This method is designed to handle the creation of a data module when
        resampling is needed during training. Instead of manually defining and
        modifying the trainer's dataloaders, this method is called to
        automatically configure the data module.

        :param float train_size: The percentage of elements to include in the
            training dataset.
        :param float test_size: The percentage of elements to include in the
            test dataset.
        :param float val_size: The percentage of elements to include in the
            validation dataset.
        :param int batch_size: The number of samples per batch to load.
        :param bool repeat: Whether to repeat the dataset data in each
            condition during training.
        :param bool automatic_batching: Whether to perform automatic batching
            with PyTorch.
        :param bool pin_memory: Whether to use pinned memory for faster data
            transfer to GPU.
        :param int num_workers: The number of worker threads for data loading.
        :param bool shuffle: Whether to shuffle the data during training.
        :raises RuntimeError: If not all conditions are sampled.
        """
        if not self.solver.problem.are_all_domains_discretised:
            error_message = "\n".join(
                [
                    f"""{" " * 13} ---> Domain {key} {
                    "sampled" if key in self.solver.problem.discretised_domains 
                    else
                    "not sampled"}"""
                    for key in self.solver.problem.domains.keys()
                ]
            )
            raise RuntimeError(
                "Cannot create Trainer if not all conditions "
                "are sampled. The Trainer got the following:\n"
                f"{error_message}"
            )
        self.data_module = PinaDataModule(
            self.solver.problem,
            train_size=train_size,
            test_size=test_size,
            val_size=val_size,
            batch_size=batch_size,
            repeat=repeat,
            automatic_batching=automatic_batching,
            num_workers=num_workers,
            pin_memory=pin_memory,
            shuffle=shuffle,
        )

    def train(self, **kwargs):
        """
        Manage the training process of the solver.

        :param dict kwargs: Additional keyword arguments. See `pytorch-lightning
            Trainer API <https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api>`_
            for details.
        """
        return super().fit(self.solver, datamodule=self.data_module, **kwargs)

    def test(self, **kwargs):
        """
        Manage the test process of the solver.

        :param dict kwargs: Additional keyword arguments. See `pytorch-lightning
            Trainer API <https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api>`_
            for details.
        """
        return super().test(self.solver, datamodule=self.data_module, **kwargs)

    @property
    def solver(self):
        """
        Get the solver.

        :return: The solver.
        :rtype: SolverInterface
        """
        return self._solver

    @solver.setter
    def solver(self, solver):
        """
        Set the solver.

        :param SolverInterface solver: The solver to set.
        """
        self._solver = solver

    @staticmethod
    def _check_input_consistency(
        solver,
        train_size,
        test_size,
        val_size,
        repeat,
        automatic_batching,
        compile,
    ):
        """
        Verifies the consistency of the parameters for the solver configuration.

        :param SolverInterface solver: The solver.
        :param float train_size: The percentage of elements to include in the
            training dataset.
        :param float test_size: The percentage of elements to include in the
            test dataset.
        :param float val_size: The percentage of elements to include in the
            validation dataset.
        :param bool repeat: Whether to repeat the dataset data in each
            condition during training.
        :param bool automatic_batching: Whether to perform automatic batching
            with PyTorch.
        :param bool compile: If ``True``, the model is compiled before training.
        """

        check_consistency(solver, SolverInterface)
        check_consistency(train_size, float)
        check_consistency(test_size, float)
        check_consistency(val_size, float)
        if repeat is not None:
            check_consistency(repeat, bool)
        if automatic_batching is not None:
            check_consistency(automatic_batching, bool)
        if compile is not None:
            check_consistency(compile, bool)

    @staticmethod
    def _check_consistency_and_set_defaults(
        pin_memory, num_workers, shuffle, batch_size
    ):
        """
        Checks the consistency of input parameters and sets default values
        for missing or invalid parameters.

        :param bool pin_memory: Whether to use pinned memory for faster data
            transfer to GPU.
        :param int num_workers: The number of worker threads for data loading.
        :param bool shuffle: Whether to shuffle the data during training.
        :param int batch_size: The number of samples per batch to load.
        """
        if pin_memory is not None:
            check_consistency(pin_memory, bool)
        else:
            pin_memory = False
        if num_workers is not None:
            check_consistency(pin_memory, int)
        else:
            num_workers = 0
        if shuffle is not None:
            check_consistency(shuffle, bool)
        else:
            shuffle = True
        if batch_size is not None:
            check_consistency(batch_size, int)
        return pin_memory, num_workers, shuffle, batch_size

    @property
    def compile(self):
        """
        Whether compilation is required or not.

        :return: ``True`` if compilation is required, ``False`` otherwise.
        :rtype: bool
        """
        return self._compile

    @compile.setter
    def compile(self, value):
        """
        Setting the value of compile.

        :param bool value: Whether compilation is required or not.
        """
        check_consistency(value, bool)
        self._compile = value



"""Module for the B-Spline model class."""

import warnings
import torch
from ..utils import check_positive_integer, check_consistency


class Spline(torch.nn.Module):
    r"""
    The univariate B-Spline curve model class.

    A univariate B-spline curve of order :math:`k` is a parametric curve defined
    as a linear combination of B-spline basis functions and control points:

    .. math::

        S(x) = \sum_{i=1}^{n} B_{i,k}(x) C_i, \quad x \in [x_1, x_m]

    where:

    - :math:`C_i \in \mathbb{R}` are the control points. These fixed points
      influence the shape of the curve but are not generally interpolated,
      except at the boundaries under certain knot multiplicities.
    - :math:`B_{i,k}(x)` are the B-spline basis functions of order :math:`k`,
      i.e., piecewise polynomials of degree :math:`k-1` with support on the
      interval :math:`[x_i, x_{i+k}]`.
    - :math:`X = \{ x_1, x_2, \dots, x_m \}` is the non-decreasing knot vector.

    If the first and last knots are repeated :math:`k` times, then the curve
    interpolates the first and last control points.


    .. note::

        The curve is forced to be zero outside the interval defined by the
        first and last knots.


    :Example:

    >>> from pina.model import Spline
    >>> import torch

    >>> knots1 = torch.tensor([0.0, 0.0, 0.0, 1.0, 2.0, 2.0, 2.0])
    >>> spline1 = Spline(order=3, knots=knots1, control_points=None)

    >>> knots2 = {"n": 7, "min": 0.0, "max": 2.0, "mode": "auto"}
    >>> spline2 = Spline(order=3, knots=knots2, control_points=None)

    >>> knots3 = torch.tensor([0.0, 0.0, 0.0, 1.0, 2.0, 2.0, 2.0])
    >>> control_points3 = torch.tensor([0.0, 1.0, 3.0, 2.0])
    >>> spline3 = Spline(order=3, knots=knots3, control_points=control_points3)
    """

    def __init__(self, order=4, knots=None, control_points=None):
        """
        Initialization of the :class:`Spline` class.

        :param int order: The order of the spline. The corresponding basis
            functions are polynomials of degree ``order - 1``. Default is 4.
        :param knots: The knots of the spline. If a tensor is provided, knots
            are set directly from the tensor. If a dictionary is provided, it
            must contain the keys ``"n"``, ``"min"``, ``"max"``, and ``"mode"``.
            Here, ``"n"`` specifies the number of knots, ``"min"`` and ``"max"``
            define the interval, and ``"mode"`` selects the sampling strategy.
            The supported modes are ``"uniform"``, where the knots are evenly
            spaced over :math:`[min, max]`, and ``"auto"``, where knots are
            constructed to ensure that the spline interpolates the first and
            last control points. In this case, the number of knots is adjusted
            if :math:`n < 2 * order`. If None is given, knots are initialized
            automatically over :math:`[0, 1]` ensuring interpolation of the
            first and last control points. Default is None.
        :type knots: torch.Tensor | dict
        :param torch.Tensor control_points: The control points of the spline.
            If None, they are initialized as learnable parameters with an
            initial value of zero. Default is None.
        :raises AssertionError: If ``order`` is not a positive integer.
        :raises ValueError: If ``knots`` is neither a torch.Tensor nor a
            dictionary, when provided.
        :raises ValueError: If ``control_points`` is not a torch.Tensor,
            when provided.
        :raises ValueError: If both ``knots`` and ``control_points`` are None.
        :raises ValueError: If ``knots`` is not one-dimensional.
        :raises ValueError: If ``control_points`` is not one-dimensional.
        :raises ValueError: If the number of ``knots`` is not equal to the sum
            of ``order`` and the number of ``control_points.``
        :raises UserWarning: If the number of control points is lower than the
            order, resulting in a degenerate spline.
        """
        super().__init__()

        # Check consistency
        check_positive_integer(value=order, strict=True)
        check_consistency(knots, (type(None), torch.Tensor, dict))
        check_consistency(control_points, (type(None), torch.Tensor))

        # Raise error if neither knots nor control points are provided
        if knots is None and control_points is None:
            raise ValueError("knots and control_points cannot both be None.")

        # Initialize knots if not provided
        if knots is None and control_points is not None:
            knots = {
                "n": len(control_points) + order,
                "min": 0,
                "max": 1,
                "mode": "auto",
            }

        # Initialization - knots and control points managed by their setters
        self.order = order
        self.knots = knots
        self.control_points = control_points

        # Check dimensionality of knots
        if self.knots.ndim > 1:
            raise ValueError("knots must be one-dimensional.")

        # Check dimensionality of control points
        if self.control_points.ndim > 1:
            raise ValueError("control_points must be one-dimensional.")

        # Raise error if #knots != order + #control_points
        if len(self.knots) != self.order + len(self.control_points):
            raise ValueError(
                f" The number of knots must be equal to order + number of"
                f" control points. Got {len(self.knots)} knots, {self.order}"
                f" order and {len(self.control_points)} control points."
            )

        # Raise warning if spline is degenerate
        if len(self.control_points) < self.order:
            warnings.warn(
                "The number of control points is smaller than the spline order."
                " This creates a degenerate spline with limited flexibility.",
                UserWarning,
            )

        # Precompute boundary interval index
        self._boundary_interval_idx = self._compute_boundary_interval()

    def _compute_boundary_interval(self):
        """
        Precompute the index of the rightmost non-degenerate interval to improve
        performance, eliminating the need to perform a search loop in the basis
        function on each call.

        :return: The index of the rightmost non-degenerate interval.
        :rtype: int
        """
        # Return 0 if there is a single interval
        if len(self.knots) < 2:
            return 0

        # Find all indices where knots are strictly increasing
        diffs = self.knots[1:] - self.knots[:-1]
        valid = torch.nonzero(diffs > 0, as_tuple=False)

        # If all knots are equal, return 0 for degenerate spline
        if valid.numel() == 0:
            return 0

        # Otherwise, return the last valid index
        return int(valid[-1])

    def basis(self, x):
        """
        Compute the basis functions for the spline using an iterative approach.
        This is a vectorized implementation based on the Cox-de Boor recursion.

        :param torch.Tensor x: The points to be evaluated.
        :return: The basis functions evaluated at x.
        :rtype: torch.Tensor
        """
        # Add a final dimension to x
        x = x.unsqueeze(-1)

        # Add an initial dimension to knots
        knots = self.knots.unsqueeze(0)

        # Base case of recursion: indicator functions for the intervals
        basis = (x >= knots[..., :-1]) & (x < knots[..., 1:])
        basis = basis.to(x.dtype)

        # One-dimensional knots case: ensure rightmost boundary inclusion
        if self._boundary_interval_idx is not None:

            # Extract left and right knots of the rightmost interval
            knot_left = knots[..., self._boundary_interval_idx]
            knot_right = knots[..., self._boundary_interval_idx + 1]

            # Identify points at the rightmost boundary
            at_rightmost_boundary = (
                x.squeeze(-1) >= knot_left
            ) & torch.isclose(x.squeeze(-1), knot_right, rtol=1e-8, atol=1e-10)

            # Ensure the correct value is set at the rightmost boundary
            if torch.any(at_rightmost_boundary):
                basis[..., self._boundary_interval_idx] = torch.logical_or(
                    basis[..., self._boundary_interval_idx].bool(),
                    at_rightmost_boundary,
                ).to(basis.dtype)

        # Iterative case of recursion
        for i in range(1, self.order):

            # Compute the denominators for both terms
            denom1 = knots[..., i:-1] - knots[..., : -(i + 1)]
            denom2 = knots[..., i + 1 :] - knots[..., 1:-i]

            # Ensure no division by zero
            denom1 = torch.where(
                torch.abs(denom1) < 1e-8, torch.ones_like(denom1), denom1
            )
            denom2 = torch.where(
                torch.abs(denom2) < 1e-8, torch.ones_like(denom2), denom2
            )

            # Compute the two terms of the recursion
            term1 = ((x - knots[..., : -(i + 1)]) / denom1) * basis[..., :-1]
            term2 = ((knots[..., i + 1 :] - x) / denom2) * basis[..., 1:]

            # Combine terms to get the new basis
            basis = term1 + term2

        return basis

    def forward(self, x):
        """
        Forward pass for the :class:`Spline` model.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        return torch.einsum(
            "...bi, i -> ...b",
            self.basis(x.as_subclass(torch.Tensor)).squeeze(-1),
            self.control_points,
        )

    @property
    def control_points(self):
        """
        The control points of the spline.

        :return: The control points.
        :rtype: torch.Tensor
        """
        return self._control_points

    @control_points.setter
    def control_points(self, control_points):
        """
        Set the control points of the spline.

        :param torch.Tensor control_points: The control points tensor. If None,
            control points are initialized to learnable parameters with zero
            initial value. Default is None.
        :raises ValueError: If there are not enough knots to define the control
            points, due to the relation: #knots = order + #control_points.
        """
        # If control points are not provided, initialize them
        if control_points is None:

            # Check that there are enough knots to define control points
            if len(self.knots) < self.order + 1:
                raise ValueError(
                    f"Not enough knots to define control points. Got "
                    f"{len(self.knots)} knots, but need at least "
                    f"{self.order + 1}."
                )

            # Initialize control points to zero
            control_points = torch.zeros(len(self.knots) - self.order)

        # Set control points
        self._control_points = torch.nn.Parameter(
            control_points, requires_grad=True
        )

    @property
    def knots(self):
        """
        The knots of the spline.

        :return: The knots.
        :rtype: torch.Tensor
        """
        return self._knots

    @knots.setter
    def knots(self, value):
        """
        Set the knots of the spline.

        :param value: The knots of the spline. If a tensor is provided, knots
            are set directly from the tensor. If a dictionary is provided, it
            must contain the keys ``"n"``, ``"min"``, ``"max"``, and ``"mode"``.
            Here, ``"n"`` specifies the number of knots, ``"min"`` and ``"max"``
            define the interval, and ``"mode"`` selects the sampling strategy.
            The supported modes are ``"uniform"``, where the knots are evenly
            spaced over :math:`[min, max]`, and ``"auto"``, where knots are
            constructed to ensure that the spline interpolates the first and
            last control points. In this case, the number of knots is inferred
            and the ``"n"`` key is ignored.
        :type value: torch.Tensor | dict
        :raises ValueError: If a dictionary is provided but does not contain
            the required keys.
        :raises ValueError: If the mode specified in the dictionary is invalid.
        """
        # If a dictionary is provided, initialize knots accordingly
        if isinstance(value, dict):

            # Check that required keys are present
            required_keys = {"n", "min", "max", "mode"}
            if not required_keys.issubset(value.keys()):
                raise ValueError(
                    f"When providing knots as a dictionary, the following "
                    f"keys must be present: {required_keys}. Got "
                    f"{value.keys()}."
                )

            # Uniform sampling of knots
            if value["mode"] == "uniform":
                value = torch.linspace(value["min"], value["max"], value["n"])

            # Automatic sampling of interpolating knots
            elif value["mode"] == "auto":

                # Repeat the first and last knots 'order' times
                initial_knots = torch.ones(self.order) * value["min"]
                final_knots = torch.ones(self.order) * value["max"]

                # Number of internal knots
                n_internal = value["n"] - 2 * self.order

                # If no internal knots are needed, just concatenate boundaries
                if n_internal <= 0:
                    value = torch.cat((initial_knots, final_knots))

                # Else, sample internal knots uniformly and exclude boundaries
                # Recover the correct number of internal knots when slicing by
                # adding 2 to n_internal
                else:
                    internal_knots = torch.linspace(
                        value["min"], value["max"], n_internal + 2
                    )[1:-1]
                    value = torch.cat(
                        (initial_knots, internal_knots, final_knots)
                    )

            # Raise error if mode is invalid
            else:
                raise ValueError(
                    f"Invalid mode for knots initialization. Got "
                    f"{value['mode']}, but expected 'uniform' or 'auto'."
                )

        # Set knots
        self.register_buffer("_knots", value.sort(dim=0).values)

        # Recompute boundary interval when knots change
        if hasattr(self, "_boundary_interval_idx"):
            self._boundary_interval_idx = self._compute_boundary_interval()



"""Module for the Equation Interface."""

from abc import ABCMeta, abstractmethod
import torch


class EquationInterface(metaclass=ABCMeta):
    """
    Abstract base class for equations.

    Equations in PINA simplify the training process. When defining a problem,
    each equation passed to a :class:`~pina.condition.condition.Condition`
    object must be either an :class:`~pina.equation.equation.Equation` or a
    :class:`~pina.equation.system_equation.SystemEquation` instance.

    An :class:`~pina.equation.equation.Equation` is a wrapper for a callable
    function, while :class:`~pina.equation.system_equation.SystemEquation`
    wraps a list of callable functions. To streamline code writing, PINA
    provides a diverse set of pre-implemented equations, such as
    :class:`~pina.equation.equation_factory.FixedValue`,
    :class:`~pina.equation.equation_factory.FixedGradient`, and many others.
    """

    @abstractmethod
    def residual(self, input_, output_, params_):
        """
        Abstract method to compute the residual of an equation.

        :param LabelTensor input_: Input points where the equation is evaluated.
        :param LabelTensor output_: Output tensor, eventually produced by a
            :class:`torch.nn.Module` instance.
        :param dict params_: Dictionary of unknown parameters, associated with a
            :class:`~pina.problem.inverse_problem.InverseProblem` instance.
        :return: The computed residual of the equation.
        :rtype: LabelTensor
        """

    def to(self, device):
        """
        Move all tensor attributes to the specified device.

        :param torch.device device: The target device to move the tensors to.
        :return: The instance moved to the specified device.
        :rtype: EquationInterface
        """
        # Iterate over all attributes of the Equation
        for key, val in self.__dict__.items():

            # Move tensors in dictionaries to the specified device
            if isinstance(val, dict):
                self.__dict__[key] = {
                    k: v.to(device) if torch.is_tensor(v) else v
                    for k, v in val.items()
                }

            # Move tensors in lists to the specified device
            elif isinstance(val, list):
                self.__dict__[key] = [
                    v.to(device) if torch.is_tensor(v) else v for v in val
                ]

            # Move tensor attributes to the specified device
            elif torch.is_tensor(val):
                self.__dict__[key] = val.to(device)

        return self
